@@ -359,3 +359,13 @@ pub fn replace_nodes(circuit: CircuitRc, map: &HashMap<HashBytes, CircuitRc>) ->
     )
     .unwrap_or(circuit)
 }
+
+#[pyfunction]
+pub fn prefix_all_names(circuit: CircuitRc, prefix: String) -> CircuitRc {
+    deep_map_unwrap(circuit, |x| {
+        if let Some(name) = x.name() {
+            return x.clone().rename(Some(prefix.clone() + name));
+        }
+        x
+    })
+}
