@@ -1,10 +1,16 @@
 use std::{cmp::Ordering, iter::zip};
 
-use anyhow::{bail, Context, Result};
+use anyhow::{anyhow, bail, Context, Result};
 use itertools::Itertools;
 use macro_rules_attribute::apply;
 use num_bigint::BigUint;
-use pyo3::prelude::*;
+use pyo3::{prelude::*, types::PyTuple};
+use rr_util::{
+    atr,
+    rearrange_spec::RInnerInts,
+    tensor_util::{IndexError, IndexTensor, TensorIndexSync},
+    util::EinsumAxes,
+};
 use rr_util::{
     opt_einsum::EinsumSpec,
     py_types::{
@@ -12,7 +18,7 @@ use rr_util::{
         ExtraPySelfOps, PyEinsumAxes, Tensor, PY_CIRCUIT_ITEMS, PY_UTILS,
     },
     pycall,
-    rearrange_spec::RearrangeSpec, // TODO: use new rearrange spec
+    rearrange_spec::{OpSize, RearrangeSpec}, // TODO: use new rearrange spec
     sv,
     tensor_util::{
         broadcast_shapes, Slice, TensorAxisIndex, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp,
@@ -21,7 +27,6 @@ use rr_util::{
         cumsum, dict_to_list, {hashmap_collect_except_duplicates, AxisInt},
     },
 };
-use rr_util::{tensor_util::IndexError, util::EinsumAxes};
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
@@ -81,7 +86,7 @@ impl CircuitNode for Einsum {
             .collect()
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(self.args.iter().map(|a| a.0.clone()))
     }
 
@@ -623,7 +628,7 @@ impl CircuitNode for Add {
             .collect()
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(self.nodes.iter().cloned())
     }
 
@@ -791,12 +796,7 @@ impl CircuitNode for Rearrange {
     circuit_node_auto_impl!("13204d30-2f12-4edd-8765-34bc8b458ef2");
 
     fn compute_shape(&self) -> Shape {
-        self.spec
-            .conform_to_input_shape(&self.node.info().shape, false)
-            .unwrap()
-            .shapes()
-            .unwrap()
-            .1
+        self.conform_to_input_shape_spec(false).shapes().unwrap().1
     }
 
     fn compute_hash(&self) -> blake3::Hasher {
@@ -820,7 +820,7 @@ impl CircuitNode for Rearrange {
             .collect()]
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(std::iter::once(self.node.clone()))
     }
 
@@ -909,15 +909,19 @@ impl Rearrange {
         )
     }
 
-    pub fn conform_to_input_shape(&self, coerce: bool) -> Result<Self> {
-        Ok(Rearrange::try_new(
+    pub fn conform_to_input_shape_spec(&self, coerce: bool) -> RearrangeSpec {
+        self.spec
+            .conform_to_input_shape(&self.node.info().shape.clone(), coerce)
+            .unwrap()
+    }
+
+    pub fn conform_to_input_shape(&self, coerce: bool) -> Self {
+        Rearrange::try_new(
             self.node.clone(),
-            self.spec
-                .conform_to_input_shape(&self.node.info().shape.clone(), coerce)
-                .unwrap(),
+            self.conform_to_input_shape_spec(coerce),
             None,
         )
-        .unwrap())
+        .unwrap()
     }
 }
 
@@ -1018,7 +1022,7 @@ impl CircuitNode for Index {
         result
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(std::iter::once(self.node.clone()))
     }
 
@@ -1064,7 +1068,7 @@ impl CircuitNode for Index {
         }
     }
 
-    fn device_dtype_extra<'a>(&'a self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + 'a> {
+    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
         Box::new(self.index.0.iter().filter_map(|x| match x {
             TensorAxisIndex::Tensor(tensor) if !tensor.shape().is_empty() => {
                 let mut out: TorchDeviceDtypeOp = TorchDeviceDtype::from_tensor(tensor).into();
@@ -1122,6 +1126,129 @@ impl Index {
         )
         .unwrap()
     }
+    #[staticmethod]
+    pub fn new_synchronized_to_start(
+        node: CircuitRc,
+        index: TensorIndexSync, // tensor index
+        name: Option<String>,
+    ) -> Result<Self> {
+        let index: TensorIndex = index.0;
+        // let broadcast_shape = broadc
+        let (tensor_part_is, nontensor_part_is) = (
+            (0..index.0.len() as u8)
+                .filter(|i| matches!(&index.0[*i as usize], &TensorAxisIndex::Tensor(_)))
+                .collect::<RInnerInts>(),
+            (0..index.0.len() as u8)
+                .filter(|i| !matches!(&index.0[*i as usize], &TensorAxisIndex::Tensor(_)))
+                .collect::<RInnerInts>(),
+        );
+        if tensor_part_is.is_empty() {
+            bail!(anyhow!(
+                "index synchronized_to_start needs to have at least one tensor axis"
+            ))
+        }
+        let rearranged_tensors_idxs_first = Rearrange::try_new(
+            node.clone(),
+            RearrangeSpec {
+                input_ints: (0..index.0.len() as u8).map(|x| sv![x]).collect(),
+                output_ints: std::iter::once(tensor_part_is.clone())
+                    .chain(nontensor_part_is.iter().map(|x| sv![*x]))
+                    .collect(),
+                int_sizes: sv![OpSize::NONE;index.0.len()],
+            },
+            name.as_ref().map(|x| format!("{}_pre_idx_flat", x)),
+        )?
+        .rc();
+        let tensors_index = TensorIndex(
+            tensor_part_is
+                .iter()
+                .map(|x| index.0[*x as usize].clone())
+                .collect(),
+        );
+        let broadcasted_tensor: IndexTensor = pycall!(
+            atr!(
+                pycall!(
+                    atr!(
+                        atr!(PY_CIRCUIT_ITEMS.circ_compiler_util, IndexUtil, raw),
+                        just_join_indices_sync,
+                        raw
+                    ),
+                    (
+                        tensors_index.clone(),
+                        Python::with_gil(|py| PyObject::from(PyTuple::new(
+                            py,
+                            tensor_part_is
+                                .iter()
+                                .map(|x| node.info().shape[*x as usize])
+                                .collect::<Vec<_>>()
+                        ))),
+                        match &tensors_index.0[0] {
+                            TensorAxisIndex::Tensor(t) => {
+                                TorchDeviceDtype::from_tensor(t).device
+                            }
+                            _ => panic!(),
+                        }
+                    ),
+                    raw
+                ),
+                flatten,
+                raw
+            ),
+            ()
+        );
+        let indexed = Index::try_new(
+            rearranged_tensors_idxs_first,
+            TensorIndex(vec![TensorAxisIndex::Tensor(broadcasted_tensor)]),
+            name.as_ref().map(|x| format!("{}_t_idx", x)),
+        )?
+        .rc();
+        let broadcasted_shape = broadcast_shapes(
+            &tensors_index
+                .0
+                .iter()
+                .map(|x| match x {
+                    TensorAxisIndex::Tensor(t) => t.shape().clone(),
+                    _ => panic!(),
+                })
+                .collect(),
+        )?;
+        let rearranged_back = Rearrange::try_new(
+            indexed,
+            RearrangeSpec {
+                input_ints: std::iter::once((0..broadcasted_shape.len() as u8).collect())
+                    .chain(
+                        (broadcasted_shape.len() as u8
+                            ..broadcasted_shape.len() as u8 + nontensor_part_is.len() as u8)
+                            .map(|x| sv![x]),
+                    )
+                    .collect(),
+                output_ints: (0..broadcasted_shape.len() as u8 + nontensor_part_is.len() as u8)
+                    .map(|x| sv![x])
+                    .collect(),
+                int_sizes: broadcasted_shape
+                    .iter()
+                    .map(|z| OpSize::from(Some(*z)))
+                    .chain(nontensor_part_is.iter().map(|_| OpSize::NONE))
+                    .collect(),
+            },
+            name.as_ref().map(|x| format!("{}_post_t_shape", x)),
+        )?
+        .rc();
+        let nontensor_index: Vec<TensorAxisIndex> = nontensor_part_is
+            .iter()
+            .map(|i| index.0[*i as usize].clone())
+            .collect();
+        Index::try_new(
+            rearranged_back,
+            TensorIndex(
+                vec![TensorAxisIndex::IDENT; broadcasted_shape.len()]
+                    .into_iter()
+                    .chain(nontensor_index)
+                    .collect(),
+            ),
+            name,
+        )
+    }
 }
 
 #[pyclass(extends=PyCircuitBase)]
@@ -1190,7 +1317,7 @@ impl CircuitNode for Concat {
         hasher
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(self.nodes.iter().cloned())
     }
 
@@ -1379,7 +1506,7 @@ impl CircuitNode for Scatter {
         hasher
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(std::iter::once(self.node.clone()))
     }
 
@@ -1656,7 +1783,7 @@ impl CircuitNode for Conv {
         hasher
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new([self.input.clone(), self.filter.clone()].into_iter())
     }
 
