@@ -25,7 +25,7 @@ macro_rules! circuit_node_auto_leaf_impl {
     ($uuid:literal) => {
         circuit_node_auto_impl!($uuid);
 
-        fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+        fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
             Box::new([].into_iter())
         }
 
@@ -86,7 +86,7 @@ impl CircuitNode for ArrayConstant {
         0
     }
 
-    fn device_dtype_extra<'a>(&'a self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + 'a> {
+    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
         Box::new(std::iter::once(
             TorchDeviceDtype::from_tensor(&self.value).into(),
         ))
@@ -374,5 +374,5 @@ impl ScalarConstant {
 fn test_nrc() {
     pyo3::prepare_freethreaded_python();
     let ex = ScalarConstant::nrc(0.0, sv![1, 2], None);
-    ex.compiler_print();
+    ex.print().unwrap();
 }
