@@ -1,4 +1,4 @@
-#![feature(map_first_last)]
+#![feature(let_chains)]
 #![allow(clippy::borrow_deref_ref)]
 use std::{
     collections::BTreeMap,
@@ -9,18 +9,19 @@ use std::{
     sync::Arc,
 };
 
-use anyhow::{Context, Result};
+use anyhow::{anyhow, bail, Context, Result};
 use circuit_utils::toposort_circuit;
 pub use computational_node::{
     flat_concat, flat_concat_back, Add, Concat, Conv, Einsum, Index, Rearrange, Scatter,
 };
 pub use constant::{ArrayConstant, ScalarConstant, Symbol};
 pub use cumulant::Cumulant;
-pub use generalfunction::{GeneralFunction, GeneralFunctionSpec, PyGFSpecShapeGetter};
+pub use generalfunction::{GeneralFunction, GeneralFunctionSpec};
+use itertools::Itertools;
 use macro_rules_attribute::apply;
 pub use module_nodes::{ModuleNode, ModuleNodeArgSpec, ModuleNodeSpec};
 use num_bigint::BigUint;
-use print::repr_circuit_deep_compiler;
+use print::PrintOptions;
 use pyo3::{
     exceptions::{self, PyValueError},
     prelude::*,
@@ -32,7 +33,7 @@ use rr_util::{
     eq_by_big_hash::EqByBigHash,
     py_types::{reduction_to_ints, use_rust_comp, PyOpAtAxes, PyShape, Tensor, PY_CIRCUIT_ITEMS},
     pycall, pycallable, python_error_exception,
-    rearrange_spec::{PyCountsAtAxes, RInnerInts, RearrangeSpec},
+    rearrange_spec::{OpSize, PyCountsAtAxes, RInnerInts, RearrangeSpec},
     sv,
     tensor_util::{check_canon_idxs, Shape, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp},
     util::{arc_ref_clone, arc_unwrap_or_clone, AsOp, AxisInt, HashBytes, HashBytesToPy},
@@ -40,7 +41,7 @@ use rr_util::{
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
 use uuid::uuid;
-pub use variable_nodes::{AutoTag, DiscreteVar, StoredCumulantVar};
+pub use variable_nodes::{DiscreteVar, StoredCumulantVar, Tag};
 
 use crate::print::print_circuit_stats;
 pub mod circuit_utils;
@@ -52,6 +53,7 @@ pub mod generalfunction;
 pub mod module_nodes;
 pub mod named_axes;
 pub mod nrc;
+pub mod opaque_iterative_matcher;
 pub mod parsing;
 pub mod print;
 pub mod variable_nodes;
@@ -165,13 +167,13 @@ pub trait CircuitNode: CircuitNodeInit {
         self.children().all(|c| c.info().can_be_sampled)
     }
 
-    fn device_dtype_extra<'a>(&'a self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + 'a> {
+    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
         Box::new(std::iter::empty())
     }
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>>;
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a>;
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_>;
 
     fn map_children_enumerate<F>(&self, f: F) -> Result<Self>
     where
@@ -326,18 +328,47 @@ pub trait CircuitNode: CircuitNodeInit {
             .unwrap_or(0usize.into())
     }
 
-    fn compiler_repr(&self) -> String
+    fn repr(&self) -> Result<String>
     where
         Self: Clone,
     {
-        repr_circuit_deep_compiler(&self.clone().c(), false, false, false)
+        PrintOptions::repr_deep(&Default::default(), self.crc())
     }
 
-    fn compiler_print(&self)
+    fn print(&self) -> Result<()>
     where
         Self: Clone,
     {
-        rr_util::python_println!("{}", self.compiler_repr())
+        rr_util::python_println!("{}", self.repr()?);
+        Ok(())
+    }
+
+    fn repru(&self) -> String
+    where
+        Self: Clone,
+    {
+        self.repr().unwrap()
+    }
+
+    fn printu(&self)
+    where
+        Self: Clone,
+    {
+        self.print().unwrap()
+    }
+
+    fn debug_repr(&self) -> Result<String>
+    where
+        Self: Clone,
+    {
+        print::debug_repr(self.crc())
+    }
+
+    fn debug_repru(&self) -> String
+    where
+        Self: Clone,
+    {
+        self.debug_repr().unwrap()
     }
 
     fn get_hash(&self) -> HashBytes {
@@ -485,6 +516,42 @@ pub trait CircuitNode: CircuitNodeInit {
             name,
         ))
     }
+    fn squeeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange>
+    where
+        Self: Clone,
+    {
+        let axes = check_canon_axes_as(
+            self.info().rank(),
+            &reduction_to_ints(Some(axes), self.info().rank()),
+        )?;
+        if axes.iter().any(|i| self.info().shape[*i as usize] != 1) {
+            bail!(anyhow!(
+                "trying to squeeze non-1 axes, shape {:?} axes {:?}",
+                &self.info().shape,
+                &axes
+            ))
+        }
+        let num_ints = self.info().rank() - axes.len();
+        let mut input_ints = sv![];
+        let mut counter = 0;
+        for i in 0..self.info().rank() as u8 {
+            if axes.iter().contains(&i) {
+                input_ints.push(sv![])
+            } else {
+                input_ints.push(sv![counter]);
+                counter += 1;
+            }
+        }
+        Rearrange::try_new(
+            self.crc(),
+            RearrangeSpec {
+                input_ints,
+                output_ints: (0..num_ints).map(|i| sv![i as u8]).collect(),
+                int_sizes: sv![OpSize::NONE;num_ints],
+            },
+            name,
+        )
+    }
     fn flatten(&self, name: Option<String>) -> Result<Rearrange>
     where
         Self: Clone,
@@ -594,7 +661,7 @@ macro_rules! circuit_node_eq_ord_debug {
 
         impl std::fmt::Debug for $type_name {
             fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-                f.write_str(&("(\n".to_owned() + &self.compiler_repr() + ")")) // TODO: compiler repr one iter!
+                f.write_str(&self.debug_repru())
             }
         }
     };
@@ -612,13 +679,42 @@ macro_rules! define_circuit_union_impl {
             )*
         }
 
+        paste::paste! {
+            #[macro_export]
+            macro_rules! [<$name:snake _map>] {
+                ($match_expr:expr; $wrap_name:ident => $e:expr) => (
+                    [<$name:snake _map_construct>]!($match_expr; ($wrap_name, __constructor_name) => $e)
+                )
+            }
+
+            #[macro_export]
+            macro_rules! [<$name:snake _map_construct>] {
+                ($match_expr:expr; ($wrap_name:ident, $constructor_name:ident) => $e:expr) => (
+                    match $match_expr {
+                    $(
+                        $name::$x($wrap_name) => {
+                            let $constructor_name = $name::$x;
+                            $e
+                        },
+                    )*
+                    }
+                )
+            }
+
+            #[macro_export]
+            macro_rules! [<on_ $name:snake _names>] {
+                ($m:ident) => (
+                    $m!( $( $x,)*);
+                )
+            }
+        }
+
         impl rr_util::eq_by_big_hash::EqByBigHash for $name{
             fn hash(&self) -> rr_util::util::HashBytes {
                 self.info().hash
             }
         }
 
-
         paste::paste! {
             #[derive(Debug, Clone, Copy, Eq, PartialEq, Hash, PartialOrd, Ord)]
             pub enum [<$name Type>] {
@@ -720,72 +816,51 @@ macro_rules! define_circuit_union_impl {
             )*
         }
 
+
+        paste::paste! {
         impl $crate::CircuitNodeInit for $name {
             fn init_info_impl(self) -> Result<Self> {
-                match self {
-                    $(
-                        Self::$x(node) => Ok(Self::$x(node.init_info()?)),
-                    )*
-                }
+                [<$name:snake _map_construct>]!(self;
+                    (node, construct) => Ok(construct(node.init_info()?)))
             }
 
             fn rename_impl(self, new_name: Option<String>) -> Self {
-                match self {
-                    $(
-                        Self::$x(node) => Self::$x(node.rename(new_name)),
-                    )*
-                }
+                [<$name:snake _map_construct>]!(self;
+                    (node, construct) => construct(node.rename(new_name)))
             }
 
             fn update_info_impl<F>(self, f: F) -> Result<Self>
             where
                 F: FnOnce(&mut $crate::CachedCircuitInfo),
             {
-                match self {
-                    $(
-                        Self::$x(node) => Ok(Self::$x(node.update_info(f)?)),
-                    )*
-                }
+                [<$name:snake _map_construct>]!(self;
+                    (node, construct) => Ok(construct(node.update_info(f)?)))
             }
         }
 
         impl $crate::CircuitNodeDefer for $name {
             #[inline] // hopefully inlined away?
             fn as_trait_obj(&self) -> &dyn $crate::CircuitNode {
-                match self {
-                    $(
-                        Self::$x(node) => node,
-                    )*
-                }
+                [<$name:snake _map>]!(self; node => node)
             }
 
             fn map_children_enumerate_impl<F>(&self, f: F) -> Result<Self>
             where
                 F: FnMut(usize,$crate::CircuitRc) -> Result<$crate::CircuitRc>,
             {
-                match self {
-                    $(
-                        Self::$x(node) => $crate::CircuitNode::map_children_enumerate(node, f).map(|v| Self::$x(v)),
-                    )*
-                }
+                [<$name:snake _map_construct>]!(self;
+                    (node, construct) => $crate::CircuitNode::map_children_enumerate(node, f).map(|v| construct(v)))
             }
 
             fn custom_c(self) -> $crate::Circuit {
-                match self {
-                    $(
-                        Self::$x(node) => node.c(),
-                    )*
-                }
+                [<$name:snake _map>]!(self; node => node.c())
             }
 
             fn custom_rc(self) -> $crate::CircuitRc {
-                match self {
-                    $(
-                        Self::$x(node) => node.rc(),
-                    )*
-                }
+                [<$name:snake _map>]!(self; node => node.rc())
             }
         }
+        }
 
         paste::paste! {
             impl $crate::CircuitNodeUnion for $name {
@@ -807,6 +882,12 @@ macro_rules! define_circuit_union_impl {
                     }
                 }
             }
+
+            impl pyo3::IntoPy<pyo3::PyObject> for $name {
+                fn into_py(self, py: pyo3::Python<'_>) -> pyo3::PyObject {
+                    [<$name:snake _map>]!(self; node => pyo3::IntoPy::into_py(node, py))
+                }
+            }
         }
 
         $(
@@ -816,20 +897,8 @@ macro_rules! define_circuit_union_impl {
                 }
             }
         )*
-
-        impl pyo3::IntoPy<pyo3::PyObject> for $name {
-            fn into_py(self, py: pyo3::Python<'_>) -> pyo3::PyObject {
-                match self {
-                    $(
-                        Self::$x(node) => pyo3::IntoPy::into_py(node, py),
-                    )*
-                }
-
-            }
-        }
     }
 }
-
 macro_rules! define_circuit {
     [$($x:ident),+ $(,)?] => {
         define_circuit_union_impl!(Circuit {$($x,)*});
@@ -871,7 +940,7 @@ define_circuit!(
     Scatter,
     Conv,
     ModuleNode,
-    AutoTag,
+    Tag,
     DiscreteVar,
     StoredCumulantVar,
     Cumulant,
@@ -889,15 +958,6 @@ macro_rules! define_circuit_union {
     [$name:ident {$($x:ident),+ $(,)?}] => {
         $crate::define_circuit_union_impl!($name {$($x,)*});
 
-        impl ::std::convert::From<$name> for $crate::Circuit {
-            fn from(item: $name) -> Self {
-                match item {
-                    $(
-                        $name::$x(node) => node.into(),
-                    )*
-                }
-            }
-        }
         impl ::std::convert::From<$crate::Circuit> for ::std::option::Option<$name> {
             fn from(item: $crate::Circuit) -> ::std::option::Option<$name> {
                 match item {
@@ -915,6 +975,12 @@ macro_rules! define_circuit_union {
             }
         }
         paste::paste! {
+            impl ::std::convert::From<$name> for $crate::Circuit {
+                fn from(item: $name) -> Self {
+                    [<$name:snake _map>]!(item; node => node.into())
+                }
+            }
+
             fn [<circuit_is_ $name:snake>](circuit: &$crate::Circuit) -> bool {
                 match circuit {
                     $(
@@ -1063,7 +1129,7 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.as_trait_obj().compute_can_be_sampled()
     }
 
-    fn device_dtype_extra<'a>(&'a self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + 'a> {
+    fn device_dtype_extra(&self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + '_> {
         self.as_trait_obj().device_dtype_extra()
     }
 
@@ -1071,7 +1137,7 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.as_trait_obj().child_axis_map()
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         self.as_trait_obj().children()
     }
 
@@ -1232,6 +1298,9 @@ pub enum ConstructError {
         input_shapes: Vec<Shape>,
     },
 
+    #[error("Passed python object isn't instance of GeneralFunctionSpec abstract class, ob={ob}")]
+    GeneralFunctionPyNotInstance { ob: PyObject },
+
     #[error("Concat requires at least one node")]
     ConcatZeroNodes {},
 
@@ -1263,6 +1332,12 @@ pub enum ConstructError {
     #[error("ModuleNode extract: not all leaves present in circuit, {subcirc:?}")]
     ModuleNodeExtractNotPresent { subcirc: CircuitRc },
 
+    #[error("spec_circuit={spec_circuit:?} missing_symbols={missing_symbols:?}")]
+    ModuleNodeSomeArgsNotPresent {
+        spec_circuit: CircuitRc,
+        missing_symbols: HashSet<Symbol>,
+    },
+
     #[error("Named axis higher than rank")]
     NamedAxisAboveRank {},
 
@@ -1407,8 +1482,7 @@ where
     where
         F: FnMut(CircuitRc) -> Result<()>,
     {
-        if !seen.contains(&circ.info().hash) {
-            seen.insert(circ.info().hash);
+        if seen.insert(circ.info().hash) {
             f(circ.clone())?;
 
             circ.children()
@@ -1422,6 +1496,20 @@ where
     recurse(circuit, &mut seen, &mut f)
 }
 
+// TODO: maybe this already exists?
+pub fn all_children(circuit: CircuitRc) -> HashSet<CircuitRc> {
+    let mut seen: HashSet<CircuitRc> = HashSet::default();
+    fn recurse(circ: CircuitRc, seen: &mut HashSet<CircuitRc>) {
+        if seen.insert(circ.clone()) {
+            for child in circ.children() {
+                recurse(child, seen)
+            }
+        }
+    }
+    recurse(circuit, &mut seen);
+    seen
+}
+
 pub fn visit_circuit_postorder<F>(circuit: CircuitRc, mut f: F)
 where
     F: FnMut(CircuitRc),
@@ -1743,8 +1831,8 @@ impl PyCircuitBase {
         use_rust_comp(&self.0, &object.0, comp_op)
     }
 
-    fn __repr__(&self) -> String {
-        self.compiler_repr(true, true, false)
+    fn __repr__(&self) -> Result<String> {
+        self.debug_repr()
     }
 
     #[getter]
@@ -1777,17 +1865,15 @@ impl PyCircuitBase {
         print_circuit_stats(&self.0)
     }
 
-    #[args(bijection = "false", shape_only_necessary = "true", arrows = "false")]
-    fn compiler_repr(&self, bijection: bool, shape_only_necessary: bool, arrows: bool) -> String {
-        repr_circuit_deep_compiler(self, bijection, shape_only_necessary, arrows)
+    #[args(options = "Default::default()")]
+    fn repr(&self, options: PrintOptions) -> Result<String> {
+        options.repr_deep((*self.0).crc())
     }
 
-    #[args(bijection = "false", shape_only_necessary = "true", arrows = "false")]
-    fn compiler_print(&self, bijection: bool, shape_only_necessary: bool, arrows: bool) {
-        rr_util::python_println!(
-            "{}",
-            self.compiler_repr(bijection, shape_only_necessary, arrows)
-        )
+    #[args(options = "Default::default()")]
+    fn print(&self, options: PrintOptions) -> Result<()> {
+        rr_util::python_println!("{}", self.repr(options)?);
+        Ok(())
     }
 
     fn numel(&self) -> BigUint {
@@ -1910,6 +1996,9 @@ impl PyCircuitBase {
     fn unsqueeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange> {
         self.expand_at_axes(axes, None, name)
     }
+    fn squeeze(&self, axes: PyOpAtAxes, name: Option<String>) -> Result<Rearrange> {
+        CircuitNode::squeeze(&self.clone().crc(), axes, name)
+    }
     fn flatten(&self, name: Option<String>) -> Result<Rearrange> {
         self.0.flatten(name)
     }
