@@ -13,13 +13,14 @@ use rustc_hash::FxHashMap as HashMap;
 use uuid::{uuid, Uuid};
 
 use crate::{
-    circuit_node_auto_impl, circuit_node_extra_impl,
+    all_children, circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_utils::replace_nodes,
     deep_map_op_context, deep_map_op_context_preorder_stoppable, deep_map_unwrap,
+    evaluate_fn_dtype_device,
     expand_node::{replace_expand_bottom_up, ExpandError},
     new_rc_unwrap,
     prelude::*,
-    CachedCircuitInfo, HashBytes, PyCircuitBase, Symbol,
+    ArrayConstant, CachedCircuitInfo, HashBytes, PyCircuitBase, Symbol,
 };
 
 #[pyclass]
@@ -44,49 +45,31 @@ impl ModuleNodeSpec {
         }
         for (spec, node) in zip(self.input_specs.iter(), nodes) {
             if node.info().rank() < spec.symbol.info().rank() {
-                // todo error messages
-                println!(
-                    "node shapes {:?}",
-                    nodes
-                        .iter()
-                        .map(|x| x.info().shape.clone())
-                        .collect::<Vec<_>>()
-                );
-                self.spec_circuit.compiler_print();
-                bail!(ExpandError::InconsistentBatches {
-                    batch_shapes: vec![spec.symbol.info().shape.clone(), node.info().shape.clone()],
-                    circuit: self.spec_circuit.clone(),
+                bail!(ExpandError::ModuleNodeRankReduced {
+                    node_rank: node.rank(),
+                    symbol_rank: spec.symbol.rank(),
+                    node_shape: node.shape().clone(),
+                    symbol_shape: spec.symbol.shape().clone(),
+                    spec_circuit: self.spec_circuit.clone()
                 });
             }
             if !spec.batchable && node.info().rank() > spec.symbol.info().rank() {
-                println!(
-                    "node shapes {:?}",
-                    nodes
-                        .iter()
-                        .map(|x| x.info().shape.clone())
-                        .collect::<Vec<_>>()
-                );
-                self.spec_circuit.compiler_print();
-                bail!(ExpandError::InconsistentBatches {
-                    batch_shapes: vec![spec.symbol.info().shape.clone(), node.info().shape.clone()],
-                    circuit: self.spec_circuit.clone(),
+                bail!(ExpandError::ModuleTriedToBatchUnbatchableInput {
+                    node_rank: node.rank(),
+                    symbol_rank: spec.symbol.rank(),
+                    spec: spec.clone(),
+                    spec_circuit: self.spec_circuit.clone()
                 });
             }
             if !spec.expandable
                 && node.info().shape[node.info().rank() - spec.symbol.info().rank()..]
                     != spec.symbol.info().shape[..]
             {
-                println!(
-                    "node shapes {:?}",
-                    nodes
-                        .iter()
-                        .map(|x| x.info().shape.clone())
-                        .collect::<Vec<_>>()
-                );
-                self.spec_circuit.compiler_print();
-                bail!(ExpandError::InconsistentBatches {
-                    batch_shapes: vec![spec.symbol.info().shape.clone(), node.info().shape.clone()],
-                    circuit: self.spec_circuit.clone(),
+                bail!(ExpandError::ModuleTriedToExpandUnexpandableInput {
+                    node_shape: node.shape().clone(),
+                    symbol_shape: spec.symbol.shape().clone(),
+                    spec: spec.clone(),
+                    spec_circuit: self.spec_circuit.clone()
                 });
             }
         }
@@ -136,8 +119,6 @@ impl ModuleNodeSpec {
         }
         let shapes = nodes.iter().map(|x| x.info().shape.clone()).collect();
         let mut expanded_shape = self.expand_shape(&shapes)?;
-        println!("expanded shape");
-        expanded_shape.compiler_print();
         let node_mapping: HashMap<HashBytes, CircuitRc> = nodes
             .iter()
             .enumerate()
@@ -176,26 +157,28 @@ impl ModuleNodeSpec {
         }
         hasher.finalize().into()
     }
-    pub fn map_circuit<F>(&self, f: F, name: Option<Option<String>>) -> Result<Self>
+    pub fn map_circuit<F>(&self, mut f: F, name: Option<Option<String>>) -> Result<Self>
     where
-        F: Fn(CircuitRc) -> Result<CircuitRc>,
+        F: FnMut(CircuitRc) -> Result<CircuitRc>,
     {
-        Ok(ModuleNodeSpec {
-            input_specs: self.input_specs.clone(),
-            name: name.unwrap_or(self.name.clone()),
-            spec_circuit: f(self.spec_circuit.clone())?,
-        })
+        ModuleNodeSpec::new(
+            f(self.spec_circuit.clone())?,
+            self.input_specs.clone(),
+            name.unwrap_or(self.name.clone()),
+        )
     }
-    pub fn map_circuit_unwrap<F>(&self, f: F, name: Option<Option<String>>) -> Self
+    pub fn map_circuit_unwrap<F>(&self, mut f: F, name: Option<Option<String>>) -> Self
     where
-        F: Fn(CircuitRc) -> CircuitRc,
+        F: FnMut(CircuitRc) -> CircuitRc,
     {
-        ModuleNodeSpec {
-            input_specs: self.input_specs.clone(),
-            name: name.unwrap_or(self.name.clone()),
-            spec_circuit: f(self.spec_circuit.clone()),
-        }
+        ModuleNodeSpec::new(
+            f(self.spec_circuit.clone()),
+            self.input_specs.clone(),
+            name.unwrap_or(self.name.clone()),
+        )
+        .unwrap()
     }
+    // TODO: add no_check_spec versions as needed!
 }
 #[pymethods]
 impl ModuleNodeSpec {
@@ -204,6 +187,31 @@ impl ModuleNodeSpec {
         spec_circuit: CircuitRc,
         input_specs: Vec<ModuleNodeArgSpec>,
         name: Option<String>,
+    ) -> Result<Self> {
+        let all_symbols = input_specs.iter().map(|x| x.symbol.crc()).collect();
+        let children = all_children(spec_circuit.clone());
+        if !children.is_superset(&all_symbols) {
+            bail!(ConstructError::ModuleNodeSomeArgsNotPresent {
+                spec_circuit,
+                missing_symbols: all_symbols
+                    .difference(&children)
+                    .map(|x| x.as_symbol_unwrap().clone())
+                    .collect(),
+            })
+        }
+
+        Ok(Self {
+            spec_circuit,
+            input_specs,
+            name,
+        })
+    }
+
+    #[staticmethod]
+    fn new_no_check_args(
+        spec_circuit: CircuitRc,
+        input_specs: Vec<ModuleNodeArgSpec>,
+        name: Option<String>,
     ) -> Self {
         Self {
             spec_circuit,
@@ -467,7 +475,7 @@ impl CircuitNode for ModuleNode {
         hasher
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(self.nodes.iter().cloned())
     }
 
@@ -492,10 +500,21 @@ impl CircuitNode for ModuleNode {
 
     fn eval_tensors(
         &self,
-        _tensors: &[rr_util::py_types::Tensor],
-        _device_dtype: &TorchDeviceDtype,
+        tensors: &[rr_util::py_types::Tensor],
+        device_dtype: &TorchDeviceDtype,
     ) -> Result<rr_util::py_types::Tensor> {
-        unimplemented!();
+        evaluate_fn_dtype_device(
+            ModuleNode::new(
+                tensors
+                    .iter()
+                    .map(|z| ArrayConstant::nrc(z.clone(), None))
+                    .collect(),
+                self.spec.clone(),
+                self.name_cloned(),
+            )
+            .expand(),
+            device_dtype.clone().into(),
+        )
     }
 }
 
