@@ -13,9 +13,9 @@ use smallvec::SmallVec as Sv;
 use uuid::Uuid;
 
 use crate::{
-    Add, ArrayConstant, AutoTag, CircuitNode, CircuitRc, Concat, Cumulant, DiscreteVar, Einsum,
+    Add, ArrayConstant, CircuitNode, CircuitRc, Concat, Cumulant, DiscreteVar, Einsum,
     GeneralFunction, Index, ModuleNode, ModuleNodeSpec, Rearrange, ScalarConstant, Scatter,
-    StoredCumulantVar, Symbol,
+    StoredCumulantVar, Symbol, Tag,
 };
 
 #[pyfunction(
@@ -26,8 +26,8 @@ use crate::{
     tensors_as_random_device_dtype = "TorchDeviceDtypeOp{device:None,dtype:None}",
     tensor_getter = "None"
 )]
-#[pyo3(name = "parse_compiler_repr_bijection")]
-pub fn parse_compiler_repr_bijection_py(
+#[pyo3(name = "parse_circuit")]
+pub fn parse_circuit_py(
     string: &str,
     module_spec_map: HashMap<String, ModuleNodeSpec>,
     reference_circuits: HashMap<String, CircuitRc>,
@@ -36,7 +36,7 @@ pub fn parse_compiler_repr_bijection_py(
     tensor_cache: Option<TensorCacheRrfs>,
 ) -> Result<CircuitRc> {
     let mut tensor_cache = tensor_cache;
-    parse_compiler_repr_bijection(
+    parse_circuit(
         string,
         module_spec_map,
         reference_circuits,
@@ -46,7 +46,7 @@ pub fn parse_compiler_repr_bijection_py(
     )
 }
 
-pub fn parse_compiler_repr_bijection(
+pub fn parse_circuit(
     string: &str,
     module_spec_map: HashMap<String, ModuleNodeSpec>,
     reference_circuits: HashMap<String, CircuitRc>,
@@ -270,7 +270,7 @@ pub fn parse_compiler_repr_bijection(
                 }
             }
             "GeneralFunction" => {
-                GeneralFunction::new_by_name(children, ps.extra.clone(), ps.name.clone())?.rc()
+                GeneralFunction::new_from_parse(children, ps.extra.clone(), ps.name.clone())?.rc()
             }
             "Index" => Index::try_new(
                 children[0].clone(),
@@ -290,7 +290,7 @@ pub fn parse_compiler_repr_bijection(
                 ps.name.clone(),
             )?
             .rc(),
-            "ModuleNode" => ModuleNode::try_new(
+            "Module" => ModuleNode::try_new(
                 children,
                 module_spec_map
                     .get(&ps.extra.clone())
@@ -307,11 +307,11 @@ pub fn parse_compiler_repr_bijection(
                 .ok_or_else(|| ParseError::CircuitRefNotFound {
                     name: ps.extra.clone(),
                 })?,
-            "AutoTag" => Uuid::from_str(&ps.extra)
+            "Tag" => Uuid::from_str(&ps.extra)
                 .map_err(|_| ParseError::InvalidUuid {
                     string: ps.extra.clone(),
                 })
-                .map(|uuid| AutoTag::nrc(children[0].clone(), uuid, ps.name.clone()))?,
+                .map(|uuid| Tag::nrc(children[0].clone(), uuid, ps.name.clone()))?,
             "Cumulant" => Cumulant::nrc(children, ps.name.clone()),
             "DiscreteVar" => {
                 DiscreteVar::try_new(children[0].clone(), children[1].clone(), ps.name.clone())?
