@@ -27,7 +27,7 @@ use crate::{
 ///   together.
 #[pyclass(extends=PyCircuitBase)]
 #[derive(Clone)]
-pub struct AutoTag {
+pub struct Tag {
     #[pyo3(get)]
     pub node: CircuitRc,
     #[pyo3(get)]
@@ -36,7 +36,7 @@ pub struct AutoTag {
     info: CachedCircuitInfo,
 }
 
-impl AutoTag {
+impl Tag {
     #[apply(new_rc)]
     pub fn new(node: CircuitRc, uuid: Uuid, name: Option<String>) -> (Self) {
         let mut out = Self {
@@ -50,15 +50,15 @@ impl AutoTag {
     }
 }
 
-impl CircuitNodeAutoName for AutoTag {
+impl CircuitNodeAutoName for Tag {
     fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| self.node.name().map(|n| format!("AutoTag {}", n)))
+        name.or_else(|| self.node.name().map(|n| format!("Tag {}", n)))
     }
 }
 
-circuit_node_extra_impl!(AutoTag);
+circuit_node_extra_impl!(Tag);
 
-impl CircuitNode for AutoTag {
+impl CircuitNode for Tag {
     circuit_node_auto_impl!("63fdc4ce-2f1b-40b3-b8b6-13991b54cbd7");
 
     fn compute_shape(&self) -> Shape {
@@ -76,7 +76,7 @@ impl CircuitNode for AutoTag {
         vec![(0..self.node.info().rank()).map(Some).collect()]
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(std::iter::once(self.node.clone()))
     }
 
@@ -99,18 +99,18 @@ impl CircuitNode for AutoTag {
 }
 
 #[pymethods]
-impl AutoTag {
+impl Tag {
     #[new]
     fn py_new(
         node: CircuitRc,
         uuid: Uuid,
         name: Option<String>,
-    ) -> PyResult<PyClassInitializer<AutoTag>> {
+    ) -> PyResult<PyClassInitializer<Tag>> {
         let out = Self::new(node, uuid, name);
         Ok(out.into_init())
     }
 
-    /// Creates a new AutoTag with a random UUID.
+    /// Creates a new Tag with a random UUID.
     #[staticmethod]
     pub fn new_with_random_uuid(node: CircuitRc, name: Option<String>) -> Self {
         Self::new(node, Uuid::new_v4(), name)
@@ -226,7 +226,7 @@ impl CircuitNode for DiscreteVar {
         ]
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new([self.values.clone(), self.probs_and_group.clone()].into_iter())
     }
 
@@ -270,9 +270,9 @@ impl DiscreteVar {
         Self::new_for_py(values, None, name)
     }
     #[staticmethod]
-    pub fn uniform_probs_and_group(size: usize, name: Option<String>) -> AutoTag {
+    pub fn uniform_probs_and_group(size: usize, name: Option<String>) -> Tag {
         // TODO: naming
-        AutoTag::new_with_random_uuid(
+        Tag::new_with_random_uuid(
             ScalarConstant::nrc(1.0 / (size as f64), sv![size], None),
             name,
         )
@@ -407,7 +407,7 @@ impl CircuitNode for StoredCumulantVar {
             .collect()
     }
 
-    fn children<'a>(&'a self) -> Box<dyn Iterator<Item = CircuitRc> + 'a> {
+    fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
         Box::new(self.cumulants.values().cloned())
     }
 
