@@ -7,8 +7,8 @@ use std::{
 use anyhow::{bail, Context, Error, Result};
 use bit_iter::BitIter;
 use circuit_base::{
-    computational_node::EinsumArgs, prelude::*, Add, AutoTag, CircuitType, Concat, DiscreteVar,
-    Einsum, GeneralFunction, Index, Rearrange, ScalarConstant, Scatter, Symbol,
+    computational_node::EinsumArgs, prelude::*, Add, CircuitType, Concat, DiscreteVar, Einsum,
+    GeneralFunction, Index, Rearrange, ScalarConstant, Scatter, Symbol, Tag,
 };
 use itertools::Itertools;
 use macro_rules_attribute::apply;
@@ -21,8 +21,7 @@ use rr_util::{
     rearrange_spec::{shape_to_op_shape, OpSize, RInts, RearrangeSpec},
     sv,
     tensor_util::{
-        check_canon_idxs, compose, IndexError, IndexTensor, Shape, Slice, TensorAxisIndex,
-        TensorIndex, USlice,
+        check_canon_idxs, compose, IndexTensor, Shape, Slice, TensorAxisIndex, TensorIndex, USlice,
     },
     union_find::UnionFind,
     unwrap,
@@ -132,7 +131,7 @@ pub fn add_deduplicate(add: &Add) -> Option<Add> {
 #[pyfunction]
 pub fn remove_add_few_input(add: &Add) -> Option<CircuitRc> {
     if add.nodes.is_empty() {
-        Some(ScalarConstant::new(0f64, add.info().shape.clone(), None).rc())
+        Some(ScalarConstant::new(0f64, add.info().shape.clone(), add.name_cloned()).rc())
     } else if add.nodes.len() == 1 {
         Some(add.nodes[0].clone())
     } else {
@@ -425,6 +424,15 @@ pub fn expand_maybe_diag(
     result
 }
 
+#[pyfunction]
+#[pyo3(name = "make_broadcast")]
+pub fn make_broadcast_py(node: CircuitRc, out_shape: Shape) -> Option<CircuitRc> {
+    match make_broadcast(&node, &out_shape) {
+        Ok(c) => Some(c),
+        Err(_) => None,
+    }
+}
+
 pub fn make_broadcast(node: &CircuitRc, out_shape: &Shape) -> Result<CircuitRc> {
     let rank_dif = out_shape.len() - node.info().rank();
     let input_ints = node
@@ -457,7 +465,7 @@ pub fn einsum_elim_identity(einsum: &Einsum) -> Option<CircuitRc> {
     {
         Some(einsum.args[0].0.clone())
     } else if einsum.args.is_empty() {
-        Some(ScalarConstant::new(1.0, sv![], None).rc())
+        Some(ScalarConstant::new(1.0, sv![], einsum.name_cloned()).rc())
     } else {
         None
     }
@@ -566,6 +574,9 @@ pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<Circuit
     if axes.is_empty() {
         return Some(circuit.clone());
     }
+    fn get_name<T: CircuitNode>(t: &T) -> Option<String> {
+        t.name_cloned().map(|x| format!("{}_rem_ax", x))
+    }
     match &***circuit {
         Circuit::ScalarConstant(scalar) => Some(
             ScalarConstant::new(
@@ -573,17 +584,19 @@ pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<Circuit
                 filter_out_idx(&scalar.info().shape, axes)
                     .into_iter()
                     .collect(),
-                scalar.name_cloned(),
+                get_name(scalar),
             )
             .rc(),
         ),
         Circuit::Rearrange(rearrange) => {
             // canonicalize first to make input 1s (), allowing us to remove them from output
-            let spec = rearrange.spec.canonicalize(true);
+            let spec = rearrange
+                .conform_to_input_shape_spec(false)
+                .canonicalize(true);
             Some(Rearrange::nrc_elim_identity(
                 rearrange.node.clone(),
                 spec.filter_out_axes(axes).unwrap(),
-                rearrange.name_cloned(),
+                get_name(rearrange),
             ))
         }
         _ => {
@@ -604,7 +617,7 @@ pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<Circuit
                         .canonicalize(true)
                         .filter_out_axes(axes)
                         .unwrap(),
-                    None,
+                    get_name(circuit),
                 ))
             } else {
                 None
@@ -621,7 +634,7 @@ pub fn generalfunction_pull_removable_axes(node: &GeneralFunction) -> Option<Cir
     let removable_axes_inp = get_removable_axes(&node.nodes[0]);
     let removable_batchable_inp: HashSet<usize> = removable_axes_inp
         .iter()
-        .filter(|x| **x < node.nodes[0].info().rank() - node.spec.num_non_batchable_output_dims)
+        .filter(|x| **x < node.nodes[0].info().rank() - node.num_non_batchable_output_dims)
         .cloned()
         .collect();
     if removable_batchable_inp.is_empty() {
@@ -867,30 +880,74 @@ pub fn add_make_broadcasts_explicit(add: &Add) -> Option<Add> {
     Some(Add::new(new_nodes, add.name_cloned()))
 }
 
-/// should return result instead
+pub fn distribute_once_raw(
+    einsum: &Einsum,
+    operand_idx: usize,
+    do_broadcasts: bool,
+    suffix: Option<String>,
+    mut call_on_sub: impl FnMut(Einsum) -> Result<CircuitRc>,
+) -> Result<Add> {
+    let mut on_sub = |node: CircuitRc, einsum: Einsum| -> Result<CircuitRc> {
+        let new_name = suffix
+            .as_ref()
+            .map(|s| node.name().map(|n_s| format!("{n_s}{s}")))
+            .flatten();
+        call_on_sub(einsum.rename(new_name))
+    };
+    match einsum.args[operand_idx].0.as_add() {
+        Some(add_node) => {
+            let mut add = add_node.clone();
+            if do_broadcasts && add.has_broadcast() {
+                add = add_make_broadcasts_explicit(&add).unwrap();
+            }
+            if add.has_broadcast() {
+                bail!(DistributeError::CouldntBroadcast { add, do_broadcasts });
+            }
+            if add.nodes.len() == 0 {
+                bail!(DistributeError::EmptyAddUnsupported {
+                    einsum: einsum.clone(),
+                    operand_idx
+                });
+            }
+            let summands = add
+                .nodes
+                .iter()
+                .map(|node| {
+                    let mut new_args = einsum.args.clone();
+                    new_args[operand_idx].0 = node.clone();
+                    on_sub(
+                        node.clone(),
+                        Einsum::new(new_args, einsum.out_axes.clone(), None),
+                    )
+                })
+                .collect::<Result<Vec<CircuitRc>>>()?;
+            Ok(Add::new(summands, einsum.name_cloned()))
+        }
+        None => bail!(DistributeError::OperandIsNotAdd {
+            einsum: einsum.clone(),
+            operand_idx,
+        }),
+    }
+}
+
 #[pyfunction]
-pub fn distribute(einsum: &Einsum, operand_idx: usize, do_broadcasts: bool) -> Option<Add> {
-    einsum.args[operand_idx].0.as_add().and_then(|add_node| {
-        let mut add = add_node.clone();
-        if do_broadcasts && add.has_broadcast() {
-            add = add_make_broadcasts_explicit(&add).unwrap();
-        }
-        if add.has_broadcast() || add.nodes.is_empty() {
-            println!("distribute failed");
-            add.compiler_print();
-            return None;
-        }
-        let summands = add
-            .nodes
-            .iter()
-            .map(|node| {
-                let mut new_args = einsum.args.clone();
-                new_args[operand_idx].0 = node.clone();
-                Einsum::nrc(new_args, einsum.out_axes.clone(), None)
-            })
-            .collect();
-        Some(Add::new(summands, None))
-    })
+pub fn distribute_once(
+    einsum: &Einsum,
+    operand_idx: usize,
+    do_broadcasts: bool,
+    suffix: Option<String>,
+) -> Result<Add> {
+    if operand_idx >= einsum.args.len() {
+        bail!(DistributeError::OperandIdxTooLarge {
+            einsum: einsum.clone(),
+            operand_idx,
+        })
+    }
+    distribute_once_raw(einsum, operand_idx, do_broadcasts, suffix, |c| Ok(c.rc()))
+}
+
+pub fn distribute_once_op(einsum: &Einsum, operand_idx: usize, do_broadcasts: bool) -> Option<Add> {
+    distribute_once_raw(einsum, operand_idx, do_broadcasts, None, |c| Ok(c.rc())).ok()
 }
 
 /// distribute all Adds that are direct children of this einsum
@@ -903,7 +960,7 @@ pub fn distribute_all(einsum: &Einsum) -> Option<Add> {
     if add_op_idxs.is_empty() {
         return None;
     }
-    let mut distributed = distribute(einsum, add_op_idxs[0], true).unwrap();
+    let mut distributed = distribute_once_op(einsum, add_op_idxs[0], true).unwrap();
     if add_op_idxs.len() > 1 {
         distributed = distributed
             .map_children(&mut |node: CircuitRc| match &**node {
@@ -940,7 +997,11 @@ pub fn einsum_of_permute_merge(einsum: &Einsum) -> Option<Einsum> {
     if !did_anything {
         return None;
     }
-    Some(Einsum::new(new_args, einsum.out_axes.clone(), None))
+    Some(Einsum::new(
+        new_args,
+        einsum.out_axes.clone(),
+        einsum.name_cloned(),
+    ))
 }
 
 #[pyfunction]
@@ -953,7 +1014,7 @@ pub fn permute_of_einsum_merge(rearrange: &Rearrange) -> Option<Einsum> {
                     .iter()
                     .map(|i| einsum.out_axes[*i])
                     .collect(),
-                None,
+                rearrange.name_cloned(),
             )
         })
     })
@@ -965,7 +1026,11 @@ pub fn einsum_elim_zero(einsum: &Einsum) -> Option<ScalarConstant> {
         Circuit::ScalarConstant(scalar) => scalar.value == 0.0,
         _ => false,
     }) {
-        Some(ScalarConstant::new(0.0, einsum.info().shape.clone(), None))
+        Some(ScalarConstant::new(
+            0.0,
+            einsum.info().shape.clone(),
+            einsum.name_cloned(),
+        ))
     } else {
         None
     }
@@ -1100,15 +1165,18 @@ pub fn index_split_axes(node: &Index, top_axes: std::collections::HashSet<usize>
 #[pyfunction]
 pub fn rearrange_fuse(node: &Rearrange) -> Option<Rearrange> {
     node.node.as_rearrange().and_then(|inner| {
-        RearrangeSpec::fuse(&inner.spec, &node.spec)
-            .map(|spec| {
-                Rearrange::new(
-                    inner.node.clone(),
-                    spec.canonicalize(true),
-                    node.name_cloned(),
-                )
-            })
-            .ok()
+        RearrangeSpec::fuse(
+            &inner.conform_to_input_shape_spec(false),
+            &node.conform_to_input_shape_spec(false),
+        )
+        .map(|spec| {
+            Rearrange::new(
+                inner.node.clone(),
+                spec.canonicalize(true),
+                node.name_cloned(),
+            )
+        })
+        .ok()
     })
 }
 
@@ -1314,18 +1382,16 @@ pub fn push_down_index_raw(
             }
         }
         Circuit::GeneralFunction(inner) => {
-            if !inner.spec.is_batchable()
-                || inner.spec.num_non_batchable_output_dims == inner.info().rank()
-            {
+            if !inner.is_batchable() || inner.num_non_batchable_output_dims == inner.info().rank() {
                 bail!(PushDownIndexError::NoopOnGeneralFunction {
                     index_node: node.clone(),
-                    is_batchable: inner.spec.is_batchable(),
-                    num_non_batch_out: inner.spec.num_non_batchable_output_dims,
+                    is_batchable: inner.is_batchable(),
+                    num_non_batch_out: inner.num_non_batchable_output_dims,
                     inner_rank: inner.rank()
                 })
             } else {
-                let rank_to_pass = inner.info().rank() - inner.spec.num_non_batchable_output_dims;
-                let new_operands = zip(&inner.nodes, &inner.spec.input_batchability)
+                let rank_to_pass = inner.info().rank() - inner.num_non_batchable_output_dims;
+                let new_operands = zip(&inner.nodes, &inner.input_batchability)
                     .map(|(child, batchable)| {
                         if *batchable {
                             let ident_rank = child.info().rank() - rank_to_pass;
@@ -1369,8 +1435,8 @@ pub fn push_down_index_raw(
                 } else {
                     bail!(PushDownIndexError::GeneralFunctionSomeAxesNotPossible {
                         index_node: node.clone(),
-                        is_batchable: inner.spec.is_batchable(),
-                        num_non_batch_out: inner.spec.num_non_batchable_output_dims,
+                        is_batchable: inner.is_batchable(),
+                        num_non_batch_out: inner.num_non_batchable_output_dims,
                         inner_rank: inner.rank(),
                         top_axis_indices
                     })
@@ -1546,7 +1612,7 @@ pub fn push_down_index_raw(
                 })
             }
         }
-        Circuit::AutoTag(tag) => AutoTag::nrc(
+        Circuit::Tag(tag) => Tag::nrc(
             on_sub(tag.node.clone(), node.index.clone(), None)?,
             tag.uuid,
             out_name,
@@ -1569,11 +1635,10 @@ pub fn push_down_index_raw(
         //     cum.dims_for_self().map(...)
         //     out_name,
         // ),
-        Circuit::Index(_) => bail!(Error::from(PushDownIndexError::UnimplementedType {
+        Circuit::Index(_) => bail!(Error::from(PushDownIndexError::ThroughIndex {
             index_node: node.clone(),
             circ_type: CircuitType::Index,
-        })
-        .context("can't push down index through other index, but you can use 'index_fuse'")),
+        })),
         inner => bail!(PushDownIndexError::UnimplementedType {
             index_node: node.clone(),
             circ_type: inner.type_tag()
@@ -1583,7 +1648,6 @@ pub fn push_down_index_raw(
     Ok(out)
 }
 
-// This is NOT Python's push_down_index, which applies recursively
 #[pyfunction]
 pub fn push_down_index_once(
     node: Index,
@@ -1930,6 +1994,11 @@ pub enum PushDownIndexError {
         out_index: TensorIndex,
         // TODO: more
     },
+    #[error("can't push down index through other index, but you can use 'index_fuse'\nindex_node={index_node:?} circ_type={circ_type:?}")]
+    ThroughIndex {
+        index_node: Index,
+        circ_type: CircuitType,
+    },
     #[error("index_node={index_node:?} circ_type={circ_type:?}")]
     UnimplementedType {
         index_node: Index,
@@ -1939,3 +2008,20 @@ pub enum PushDownIndexError {
     #[error("tensors aren't currently supported tensor={tensor:?}")]
     ConcatSplitSectionsTensorUnsupported { tensor: IndexTensor },
 }
+
+#[apply(python_error_exception)]
+#[base_error_name(Distribute)]
+#[base_exception(PyValueError)]
+#[derive(Error, Debug, Clone)]
+pub enum DistributeError {
+    #[error("add={add:?} do_broadcasts={do_broadcasts}")]
+    CouldntBroadcast { add: Add, do_broadcasts: bool },
+    #[error("einsum={einsum:?} operand_idx={operand_idx}")]
+    OperandIsNotAdd { einsum: Einsum, operand_idx: usize },
+    #[error("einsum={einsum:?} operand_idx={operand_idx}")]
+    OperandIdxTooLarge { einsum: Einsum, operand_idx: usize },
+    #[error("einsum={einsum:?} operand_idx={operand_idx}")]
+    EmptyAddUnsupported { einsum: Einsum, operand_idx: usize },
+    #[error("einsum={einsum:?} operand_idx={operand_idx}")]
+    Noop { einsum: Einsum, operand_idx: usize },
+}
