@@ -40,11 +40,11 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
     // println!("leaves axis");
     // for (k, v) in &leaves_axis {
     //     println!("{}", v);
-    //     k.compiler_print();
+    //     k.print();
     // }
     // println!("leaves non axis");
     // for k in &leaves_non_axis {
-    //     k.compiler_print();
+    //     k.print();
     // }
     let input_specs_axis: Vec<(CircuitRc, ModuleNodeArgSpec)> = leaves_axis
         .iter()
@@ -78,7 +78,7 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
         .collect();
     // println!("input_specs");
     // for is in &input_specs {
-    //     is.0.compiler_print();
+    //     is.0.print();
     // }
     if input_specs.is_empty() {
         bail!(BatchError::AxisOriginatesTooHigh {});
@@ -86,7 +86,7 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
     let module_spec =
         ModuleNodeSpec::new_extract(circuit.clone(), input_specs.clone(), None, false)?;
     // println!("module spec");
-    // module_spec.spec_circuit.compiler_print();
+    // module_spec.spec_circuit.print();
     let module_spec = module_spec.resize(
         module_spec
             .input_specs
@@ -105,7 +105,7 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
             .collect(),
     )?;
     // println!("module spec after resize");
-    // module_spec.spec_circuit.compiler_print();
+    // module_spec.spec_circuit.print();
     let concattands: Vec<CircuitRc> = (0..num_batches)
         .map(|i| {
             let new_nodes: Vec<CircuitRc> = module_spec
@@ -140,7 +140,7 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
 
             // println!("module args");
             // for c in new_nodes.iter() {
-            //     c.compiler_print();
+            //     c.print();
             // }
             ModuleNode::nrc(new_nodes, module_spec.clone(), None)
         })
@@ -153,9 +153,9 @@ pub fn batch_to_concat(circuit: CircuitRc, axis: usize, num_batches: usize) -> R
             circuit.info().shape
         );
         println!("old");
-        circuit.compiler_print();
+        circuit.printu();
         println!("new");
-        result.compiler_print();
+        result.printu();
         panic!();
     }
     Ok(result)
@@ -219,7 +219,6 @@ pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> Resul
     let nested_big =
         einsum_nest_optimize(&einsum_named_axes, &mut no_memory_limit_context).unwrap();
 
-    nested_big.compiler_print();
     let mut intermediates_above_mem = vec![];
     visit_circuit_unwrap(nested_big.crc(), |c| {
         if c.info().numel_usize() > max_single_mem {
