@@ -1,9 +1,9 @@
 use std::io::prelude::*;
 
-use anyhow::{bail, Result};
+use anyhow::{bail, Context, Result};
 use circuit_base::{
     circuit_utils::has_leaf_symbols, deep_map_op_context, flat_concat, get_compatible_dtype,
-    print::repr_circuit_deep_compiler, CircuitNode, CircuitRc, ModuleNode, ModuleNodeSpec,
+    print::PrintOptions, CircuitNode, CircuitRc, ModuleNode, ModuleNodeSpec,
 };
 use num_bigint::BigUint;
 use pyo3::prelude::*;
@@ -87,7 +87,7 @@ impl Default for OptimizationSettings {
             distribute_min_size: None,
             scheduling_naive: false,
             scheduling_timeout: 5_000,
-            scheduling_simplify: false,
+            scheduling_simplify: true,
             scheduling_simplify_lossy: false,
             adjust_numerical_scale: false,
             numerical_scale_min: 1e-8,
@@ -249,10 +249,13 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 4 {
         python_println!("Original Circuit");
-        circuit.compiler_print();
+        circuit.printu();
     }
     if context.settings.save_circ {
-        let s = repr_circuit_deep_compiler(&circuit, true, true, false);
+        let s = PrintOptions::new(true, true, None, false, false)
+            .unwrap()
+            .repr_deep(circuit.clone())
+            .expect("can't fail with None traversal");
         let f_name = format!("cached_circ_{}.circ", circuit.info().hash_usize());
         python_println!("Save circ to {}", f_name);
 
@@ -274,7 +277,7 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Simplified Circuit");
-        circuit.compiler_print();
+        circuit.printu();
     }
     let circuit = timed!(
         apply_fn_until_same(&circuit, |x| {
@@ -288,7 +291,7 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Distributed Circuit");
-        circuit.compiler_print();
+        circuit.printu();
     }
     let circuit = timed!(deep_heuristic_nest_adds(circuit), 10, print_timings);
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
@@ -298,7 +301,7 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Final Circuit");
-        circuit.compiler_print();
+        circuit.printu();
     }
     circuit
 }
@@ -314,10 +317,10 @@ pub fn optimize_to_schedule(
         python_println!("Optimizing verbose {}", verbose)
     }
     if settings.capture_and_print {
-        python_println!(
-            "{}",
-            repr_circuit_deep_compiler(&**circuit, true, false, false)
-        );
+        PrintOptions::new(true, false, None, false, false)
+            .unwrap()
+            .print(circuit.clone())
+            .context("capture print fail!")?;
     }
     let optimized_circuit = timed!(optimize_circuit(circuit, context), 10, verbose >= 1);
     // return evaluate_fn(&*optimized_circuit).unwrap();
