@@ -143,8 +143,8 @@ pub fn add_pull_concat(add: &Add) -> Option<CircuitRc> {
                     .rc();
                 let concat_of_adds = split_to_concat(add_explicit, concat_axis, sections_here).rc();
                 if add.info().shape[..] != concat_of_adds.info().shape[..] {
-                    add.crc().compiler_print();
-                    concat_of_adds.compiler_print();
+                    add.crc().printu();
+                    concat_of_adds.printu();
                     panic!();
                 }
                 return Some(concat_of_adds);
@@ -161,8 +161,7 @@ pub fn add_pull_concat(add: &Add) -> Option<CircuitRc> {
 // pub fn rearrange_pull_concat(rearrange: &Rearrange) -> Option<Concat> {}
 #[pyfunction]
 pub fn generalfunction_pull_concat(generalfunction: &GeneralFunction) -> Option<Concat> {
-    let batch_rank =
-        generalfunction.info().rank() - generalfunction.spec.num_non_batchable_output_dims;
+    let batch_rank = generalfunction.info().rank() - generalfunction.num_non_batchable_output_dims;
     for operand in &generalfunction.nodes {
         if let Circuit::Concat(concat) = &***operand &&  concat.axis < batch_rank {
             let mut cur = 0;
@@ -173,7 +172,7 @@ pub fn generalfunction_pull_concat(generalfunction: &GeneralFunction) -> Option<
                     let span_here = cnode.info().shape[concat.axis];
                     let new_operands = zip(
                         &generalfunction.nodes,
-                        &generalfunction.spec.input_batchability,
+                        &generalfunction.input_batchability,
                     )
                     .map(|(node, batchability)| {
                         if !batchability {
