@@ -79,12 +79,12 @@ pub fn opt_eval_each_subcircuit_until_fail(circuit: CircuitRc, settings: Optimiz
                 let close_result = is_close(reference, optimized);
                 if close_result.is_err() {
                     println!("{:?}", close_result);
-                    circ.compiler_print();
+                    circ.printu();
                     optimize_circuit(
                         circ,
                         &mut OptimizationContext::new_settings_circuit(settings, circuit.clone()),
                     )
-                    .compiler_print();
+                    .printu();
                 }
             }
         }
