@@ -1,13 +1,14 @@
 use circuit_base::{
-    evaluate_fn_dtype_device, generalfunction::OFFICIAL_GENERALFUNCTION_INVERSES, Circuit,
-    CircuitNode, CircuitRc, GeneralFunction, ScalarConstant,
+    evaluate_fn_dtype_device,
+    generalfunction::{SpecTrait, OFFICIAL_GENERALFUNCTION_INVERSES},
+    Circuit, CircuitNode, CircuitRc, GeneralFunction, ScalarConstant,
 };
 use pyo3::prelude::*;
 use rr_util::{sv, tensor_util::TorchDeviceDtypeOp};
 
 #[pyfunction]
 pub fn generalfunction_merge_inverses(circuit: &GeneralFunction) -> Option<CircuitRc> {
-    if circuit.nodes.len()==1 && circuit.spec.is_official && let Circuit::GeneralFunction(inner)=&**circuit.nodes[0] && inner.spec.is_official && OFFICIAL_GENERALFUNCTION_INVERSES.iter().any(|(a,b)|*a==&circuit.spec.name&&*b==&inner.spec.name) {
+    if circuit.nodes.len()==1 && circuit.spec.is_official() && let Circuit::GeneralFunction(inner)=&**circuit.nodes[0] && inner.spec.is_official() && OFFICIAL_GENERALFUNCTION_INVERSES.iter().any(|(a,b)|*a==&circuit.spec.name()&&*b==&inner.spec.name()) {
         return Some(inner.nodes[0].clone())
     }
     None
@@ -15,8 +16,8 @@ pub fn generalfunction_merge_inverses(circuit: &GeneralFunction) -> Option<Circu
 
 #[pyfunction]
 pub fn generalfunction_special_case_simplification(circuit: &GeneralFunction) -> Option<CircuitRc> {
-    if circuit.spec.is_official {
-        let name_str: &str = &circuit.spec.name;
+    if circuit.spec.is_official() {
+        let name_str: &str = &circuit.spec.name();
         match name_str {
             "softmax" => {
                 if let Circuit::ScalarConstant(_sc) = &**circuit.nodes[0] {
@@ -50,7 +51,7 @@ pub fn generalfunction_special_case_simplification(circuit: &GeneralFunction) ->
 /// this does less than python version, maybe should expand
 #[pyfunction]
 pub fn generalfunction_evaluate_simple(node: &GeneralFunction) -> Option<CircuitRc> {
-    if node.spec.num_non_batchable_output_dims == 0
+    if node.num_non_batchable_output_dims == 0
         && node.info().rank() == 0
         && node.nodes.len() == 1
         && node.nodes[0].info().rank() == 0
