@@ -12,17 +12,16 @@ use circuit_base::{
 };
 use itertools::Itertools;
 use macro_rules_attribute::apply;
-use miniserde::json::{self, Array};
+use miniserde::json;
 use num_bigint::BigUint;
 use pyo3::{exceptions::PyValueError, prelude::*, types::PyBytes};
 use rr_util::{
-    cached_lambda,
     lru_cache::TensorCacheRrfs,
     py_types::{scalar_to_tensor, tensor_scale, un_flat_concat, ExtraPySelfOps, Tensor, PY_UTILS},
-    pycall, python_error_exception,
+    pycall, python_error_exception, sv,
     tensor_util::{Shape, TorchDeviceDtype},
     timed,
-    util::{arc_ref_clone, filter_to_idx, HashBytes},
+    util::{arc_ref_clone, with_context_failable, HashBytes},
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
@@ -30,7 +29,7 @@ use thiserror::Error;
 use crate::{
     circuit_optimizer::{OptimizationContext, OptimizationSettings},
     schedule_send::ScheduleToSend,
-    scheduling_z3::schedule_dag_strategy_ints,
+    scheduling_alg::{Dag, DagSimpSettings},
 };
 
 #[derive(Clone, Debug)]
@@ -188,9 +187,9 @@ impl Schedule {
             + 1
     }
 
-    pub fn serialize(&self) -> String {
-        let tosend: ScheduleToSend = self.into();
-        json::to_string(&tosend)
+    pub fn serialize(&self) -> Result<String> {
+        let tosend: ScheduleToSend = self.try_into()?;
+        Ok(json::to_string(&tosend))
     }
 
     #[staticmethod]
@@ -204,17 +203,21 @@ impl Schedule {
         Schedule::deserialize(string, device, &mut tensor_cache)
     }
 
-    pub fn tosend(&self) -> ScheduleToSend {
-        self.into()
+    pub fn tosend(&self) -> Result<ScheduleToSend> {
+        self.try_into()
     }
 
-    pub fn evaluate_remote(&self, remote_url: String) -> Option<Tensor> {
-        let tosend: ScheduleToSend = self.into();
-        tosend.evaluate_remote(remote_url, self.device_dtype.device.clone())
+    pub fn evaluate_remote(&self, remote_url: String) -> Result<Option<Tensor>> {
+        let out = self
+            .tosend()?
+            .evaluate_remote(remote_url, self.device_dtype.device.clone());
+        Ok(out)
     }
-    pub fn evaluate_remote_many(&self, remote_url: String) -> Option<Vec<Tensor>> {
-        let tosend: ScheduleToSend = self.into();
-        tosend.evaluate_remote_many(remote_url, self.device_dtype.device.clone())
+    pub fn evaluate_remote_many(&self, remote_url: String) -> Result<Option<Vec<Tensor>>> {
+        let out = self
+            .tosend()?
+            .evaluate_remote_many(remote_url, self.device_dtype.device.clone());
+        Ok(out)
     }
 }
 impl Schedule {
@@ -393,7 +396,7 @@ pub fn evaluate_schedule(schedule: &Schedule) -> HashMap<usize, Tensor> {
                 let result_err = circ.eval_tensors(&tensors, &schedule.device_dtype);
                 if result_err.is_err() {
                     println!("errored evaluate");
-                    circ.compiler_print()
+                    circ.print().unwrap()
                 }
                 let result = result_err.unwrap();
                 assert!(result.shape()[..] == circ.info().shape[..]);
@@ -628,225 +631,49 @@ pub fn order_to_schedule(
     (result, to_keep.iter().map(|x| circ_to_id[x]).collect())
 }
 
-#[pyclass]
-#[derive(Default, Clone, Debug)]
-pub struct Dag {
-    pub children: Vec<Vec<usize>>,
-    pub parents: Vec<Vec<usize>>,
-    /// the cost is the number of new elements computing the node entails allocating.
-    /// some nodes compute views into existing memory, such as some indexes and all ArrayConstants
-    /// (as they're already allocated), so they have cost 0
-    /// if a node computes a new tensor, the cost is the product of the shape.
-    pub node_costs: Vec<usize>,
-    pub node_hashes: Vec<HashBytes>,
-    pub hash_to_node: HashMap<HashBytes, usize>,
-}
-impl Dag {
-    pub fn get_outputs(&self) -> Vec<usize> {
-        let result = filter_to_idx(&self.parents.iter().collect(), &mut |y| y.is_empty());
-        result
-    }
-    /// Create a new dag with a subset of the nodes. Edges to ignored
-    /// nodes are replaced with edges that pass through to nodes we're keeping
-    pub fn sub_dag(&self, nodes_to_keep: &Vec<usize>) -> Dag {
-        // map indices from nodes_to_keep -> arange
-        let old_to_new: HashMap<usize, usize> = nodes_to_keep
-            .iter()
-            .enumerate()
-            .map(|x| (*x.1, x.0))
-            .collect();
-        // fill out everything except the parents and children
-        // assumes the costs stay the same
-        let mut result = Dag {
-            node_hashes: nodes_to_keep.iter().map(|x| self.node_hashes[*x]).collect(),
-            hash_to_node: nodes_to_keep
-                .iter()
-                .enumerate()
-                .map(|(i, x)| (self.node_hashes[*x], i))
-                .collect(),
-            node_costs: nodes_to_keep.iter().map(|x| self.node_costs[*x]).collect(),
-            children: vec![vec![]; nodes_to_keep.len()],
-            parents: vec![vec![]; nodes_to_keep.len()],
-        };
-        // new children = (old children we're keeping) + (new children of old children we're not keeping)
-        #[apply(cached_lambda)]
-        fn new_children(old_i: usize) -> Vec<usize> {
-            let old_children = &self.children[old_i];
-            old_children
-                .iter()
-                .flat_map(|old_child| {
-                    if let Some(n) = old_to_new.get(old_child) {
-                        vec![*n]
-                    } else {
-                        new_children(*old_child)
-                    }
-                })
-                .collect()
-        }
-        // fill out parents and children
-        for (new_i, old_i) in nodes_to_keep.iter().enumerate() {
-            let children = new_children(*old_i);
-            result.children[new_i] = children.clone();
-            for child in children {
-                result.parents[child].push(new_i);
-            }
-        }
-        result
-    }
-
-    /// Find all of the nontrivial chains within the dag which are internally not connected
-    /// to other nodes, and return the top / parent indices
-    /// This corresponds to all the size >=2 subcircuits in a circuit which involve serial computation
-    /// (Technically these are the nontrivial maximal chain induced subgraphs)
-    pub fn chains(&self) -> Vec<usize> {
-        #[apply(cached_lambda)]
-        fn top_of_chain(node: usize) -> usize {
-            if self.parents[node].len() == 1 && self.children[self.parents[node][0]].len() == 1 {
-                top_of_chain(self.parents[node][0])
-            } else {
-                node
-            }
-        }
-        let node_to_top: Vec<usize> = (0..self.node_costs.len()).map(top_of_chain).collect();
-        node_to_top
-            .into_iter()
-            .enumerate()
-            // filter out length 1 chains
-            .filter_map(|(node, top)| if node == top { None } else { Some(top) })
-            // return each parent of a nontrivial chain once
-            .unique()
-            .collect()
-    }
-}
-
 pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
     let mut result: Dag = Default::default();
     // append the circuit to the dag if it's not already there, and return its index
-    let number_node = |c: CircuitRc, result: &mut Dag| -> usize {
+    let number_node = |c: CircuitRc, result: &mut Dag| -> u32 {
         if let Some(idx) = result.hash_to_node.get(&c.info().hash) {
-            return *idx;
+            return *idx as u32;
         }
         result.node_costs.push(c.intermediate_cost_bound());
         result.node_hashes.push(c.info().hash);
         result
             .hash_to_node
             .insert(c.info().hash, result.node_hashes.len() - 1);
-        result.children.push(vec![]);
-        result.parents.push(vec![]);
-        result.node_hashes.len() - 1
+        result.children.push(sv![]);
+        result.parents.push(sv![]);
+        (result.node_hashes.len() - 1) as u32
     };
     for circuit in circuits {
         visit_circuit_unwrap(circuit.clone(), |c: CircuitRc| {
             // ArrayConstants are never added to the dag to begin with because they're always 0 cost
             if !Leaf::matches(&c) {
-                let my_number = number_node(c.clone(), &mut result);
+                let my_number: u32 = number_node(c.clone(), &mut result);
                 let children_to_consider: Vec<CircuitRc> =
                     c.children().filter(|child| !Leaf::matches(child)).collect();
-                result.children[my_number] = children_to_consider
+                result.children[my_number as usize] = children_to_consider
                     .iter()
                     .map(|child| number_node(child.clone(), &mut result))
+                    .unique()
                     .collect();
                 for child in children_to_consider {
                     let new_num = number_node(child.clone(), &mut result);
-                    result.parents[new_num].push(my_number);
+                    if !result.parents[new_num as usize]
+                        .iter()
+                        .contains(&(my_number as u32))
+                    {
+                        result.parents[new_num as usize].push(my_number as u32);
+                    }
                 }
             }
         });
     }
-    result
-}
-
-fn simplify_dag_for_scheduling_and_kept_nodes(
-    dag: &Dag,
-    chunk_size: usize,
-    lossy: bool,
-) -> (Dag, Vec<usize>) {
-    // filter out nodes that are too small to be worth scheduling (such as cost 0 indexes)
-    let mut dag = dag.clone();
-    if lossy {
-        let nodes_to_keep: Vec<usize> = (0..dag.node_costs.len())
-            .filter(|i| dag.node_costs[*i] >= chunk_size)
-            .collect();
-        dag = dag.sub_dag(&nodes_to_keep);
-    }
-
-    // now we filter out some parts in chains of serial computation
-    // if we have a chain of nodes `start -> node_1 -> ... -> end`,
-    // we can simplify the dag to only include the node with the smallest cost,
-    // and the highest cost nodes on either side.
-    // we need the smallest cost node as after computing this is the best time to compute
-    // other nodes from other places in parallel
-    // and we need the maximum nodes on either side so the scheduler has the right bound
-    // on intermediate memory consumption along the serial chain
-    let chain_ends = dag.chains();
-    let mut nodes_to_prune: HashSet<usize> = HashSet::default();
-    for chain_end in chain_ends {
-        // iterate backwards down the chain and accumulate the min and max on either side
-        let c = &dag.node_costs;
-        let mut current = (chain_end, c[chain_end]);
-        let mut min_node = current;
-        let mut max_before = current;
-        let mut max_after = current;
-        while dag.children[current.0].len() == 1 {
-            current = {
-                let child = dag.children[current.0][0];
-                (child, c[child])
-            };
-            if current.1 < min_node.1 {
-                max_before = *vec![max_before, min_node, max_after, current]
-                    .iter()
-                    .max_by_key(|(_node, cost)| cost)
-                    .unwrap();
-                min_node = current;
-                max_after = current;
-            } else if current.1 > max_after.1 {
-                max_after = current;
-            }
-        }
-        // and drop all the other nodes
-        let mut current = chain_end;
-        while dag.children[current].len() == 1 {
-            current = dag.children[current][0];
-            if current != min_node.0 && current != max_before.0 && current != max_after.0 {
-                nodes_to_prune.insert(current);
-            }
-        }
-    }
-    let nodes_to_keep = (0..dag.node_costs.len())
-        .filter(|i| !nodes_to_prune.contains(i))
+    result.node_to_orig = (0..result.node_costs.len() as u32)
+        .map(|x| (x, sv![x].clone()))
         .collect();
-    let dag = dag.sub_dag(&nodes_to_keep);
-    (dag, nodes_to_keep)
-}
-
-fn subset_order_to_full_order(
-    order: &Vec<usize>,
-    kept_nodes: &Vec<usize>,
-    dag: &Dag,
-) -> Vec<usize> {
-    let mut result: Vec<usize> = vec![];
-    let mut result_set: HashSet<usize> = HashSet::default();
-
-    fn recurse(orig_i: usize, result: &mut Vec<usize>, result_set: &mut HashSet<usize>, dag: &Dag) {
-        if !result_set.contains(&orig_i) {
-            for child in &dag.children[orig_i] {
-                if !result_set.contains(child) {
-                    recurse(*child, result, result_set, dag);
-                }
-            }
-            result.push(orig_i);
-            result_set.insert(orig_i);
-        }
-    }
-
-    for sub_i in order {
-        let orig_i = kept_nodes[*sub_i];
-        recurse(orig_i, &mut result, &mut result_set, dag)
-    }
-    for output in dag.get_outputs() {
-        recurse(output, &mut result, &mut result_set, dag);
-    }
-
     result
 }
 
@@ -854,14 +681,14 @@ pub fn circuit_to_schedule(
     circuit: CircuitRc,
     context: &mut OptimizationContext,
 ) -> Result<Schedule> {
-    let dag = circuits_to_dag(&[circuit.clone()]);
+    let mut dag = circuits_to_dag(&[circuit.clone()]);
     if circuit.info().max_non_input_size > BigUint::from(context.cache.fallback_total_numel) {
         bail!(SchedulingOOMError::Single {
             numel: circuit.info().max_non_input_size.clone(),
         });
     }
     let order_result = {
-        let mut result = Err(SchedulingOOMError::Many {
+        let mut result: Result<Vec<u32>, SchedulingOOMError> = Err(SchedulingOOMError::Many {
             max_memory: 0,
             memory_chunks: 0,
             node_memories: vec![],
@@ -870,51 +697,39 @@ pub fn circuit_to_schedule(
         let mut mem = context.cache.max_tensor_elements;
         while mem <= context.cache.fallback_total_numel && result.is_err() {
             if context.settings.scheduling_simplify {
-                let chunk_size = mem / context.settings.scheduling_num_mem_chunks;
-                // operate on subset dag
-                let (sub_dag, kept_nodes) = simplify_dag_for_scheduling_and_kept_nodes(
-                    &dag,
-                    chunk_size,
-                    context.settings.scheduling_simplify_lossy,
-                );
-                if context.settings.verbose >= 2 {
-                    println!(
-                        "dag sizes: orig: {} pruned: {}",
-                        dag.node_costs.len(),
-                        sub_dag.node_costs.len()
-                    );
-                }
-                result = schedule_dag_strategy_ints(
-                    &sub_dag,
-                    context.settings.verbose,
-                    mem,
-                    context.settings.scheduling_num_mem_chunks,
-                    context.settings.scheduling_timeout,
-                )
-                .map(|sub_order| subset_order_to_full_order(&sub_order, &kept_nodes, &dag));
-            } else {
-                result = schedule_dag_strategy_ints(
-                    &dag,
-                    context.settings.verbose,
-                    mem,
-                    context.settings.scheduling_num_mem_chunks,
-                    context.settings.scheduling_timeout,
+                let mut dag_simp_settings: DagSimpSettings = Default::default();
+                dag_simp_settings.acceptable_subdag_max = context.cache.max_single_tensor_numel;
+                dag_simp_settings.mem_limit = mem;
+                dag_simp_settings.num_mem_chunks = context.settings.scheduling_num_mem_chunks;
+                dag_simp_settings.timeout = context.settings.scheduling_timeout;
+                dag_simp_settings.verbose = context.settings.verbose;
+                timed!(
+                    dag.simplify(&dag_simp_settings),
+                    10,
+                    context.settings.verbose >= 2
                 );
             }
+            result = dag.compute_schedule(
+                context.settings.verbose,
+                mem,
+                context.settings.scheduling_num_mem_chunks,
+                context.settings.scheduling_timeout,
+            );
             mem *= 2;
         }
         result
     };
-    if order_result.is_err() {
-        circuit.compiler_print();
-    }
-
-    let order = order_result?;
+    let order = with_context_failable(order_result, || {
+        Ok(format!(
+            "getting schedule failed for circuit:\n{}",
+            circuit.repr()?
+        ))
+    })?;
 
     let to_node = hash_to_node(circuit.clone());
     let circuit_order: Vec<CircuitRc> = order
         .iter()
-        .map(|x| to_node[&dag.node_hashes[*x]].clone())
+        .map(|x| to_node[&dag.node_hashes[*x as usize]].clone())
         .collect();
     let (mut out, kept_keys) = order_to_schedule(
         &circuit_order,
@@ -950,10 +765,15 @@ pub fn circuit_to_schedule_naive_toposort(circuit: CircuitRc) -> Schedule {
 
 #[pyfunction]
 pub fn scheduled_evaluate(circuit: CircuitRc, settings: OptimizationSettings) -> Result<Tensor> {
-    let schedule = circuit_to_schedule(
-        circuit.clone(),
-        &mut OptimizationContext::new_settings_circuit(settings, circuit),
-    )?;
+    if settings.scheduling_naive {}
+    let schedule = if settings.scheduling_naive {
+        Ok(circuit_to_schedule_naive_toposort(circuit))
+    } else {
+        circuit_to_schedule(
+            circuit.clone(),
+            &mut OptimizationContext::new_settings_circuit(settings, circuit),
+        )
+    }?;
     Ok(evaluate_schedule(&schedule)[&schedule.output_circuit.unwrap().0].clone())
 }
 
