@@ -22,6 +22,7 @@ use rr_util::{
 use uuid::uuid;
 
 use crate::{
+    library::{apply_in_traversal, replace_outside_traversal_symbols},
     BoundGetter, BoundUpdater, Getter, Matcher, MatcherData, MatcherFromPyBase, MatcherRc,
     Transform, TransformRc, Updater,
 };
@@ -499,6 +500,7 @@ impl IterativeMatcher {
         Ok(res)
     }
 
+    #[pyo3(name = "validate_matched")]
     fn validate_matched_py(&self, matched: HashSet<CircuitRc>) -> Result<()> {
         self.validate_matched(&matched)
     }
@@ -753,4 +755,21 @@ impl IterativeMatcher {
         )
         .bind(self_.crc())
     }
+
+    pub fn apply_in_traversal(
+        &self,
+        circuit: CircuitRc,
+        transform: TransformRc,
+    ) -> Result<CircuitRc> {
+        apply_in_traversal(circuit, self_.clone().rc(), |x| transform.run(x))
+    }
+
+    pub fn traversal_edges(&self, circuit: CircuitRc) -> Result<Vec<CircuitRc>> {
+        Ok(
+            replace_outside_traversal_symbols(circuit, self_.clone().rc(), |_| Ok(None))?
+                .1
+                .into_values()
+                .collect(),
+        )
+    }
 }
