@@ -1,6 +1,8 @@
+#![feature(let_chains)]
 //! Basic TODO: add more rust helpers + builders as needed!
 
 pub mod iterative_matcher;
+pub mod library;
 pub mod matcher;
 pub mod operations;
 pub mod sample_transform;
