@@ -17,7 +17,7 @@ use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     cached_method,
-    py_types::{Tensor, PY_UTILS},
+    py_types::{PyShape, Tensor, PY_UTILS},
     python_error_exception,
     tensor_util::{Shape, Slice, TensorAxisIndex, TensorIndex},
     util::HashBytes,
@@ -113,8 +113,8 @@ macro_rules! py_forward_sample_spec {
                 self.get_overall_weight()
             }
             #[pyo3(name = "get_sample_shape")]
-            fn get_sample_shape_py(&self) -> Shape {
-                self.get_sample_shape()
+            fn get_sample_shape_py(&self) -> PyShape {
+                PyShape(self.get_sample_shape())
             }
             #[pyo3(name = "get_transform")]
             fn get_transform_py(&self) -> TransformRc {
