@@ -18,17 +18,19 @@ pub fn default_index_traversal() -> IterativeMatcherRc {
         .rc()
 }
 
-#[pyfunction]
+#[pyfunction(
+    traversal = "default_index_traversal()",
+    suffix = "None",
+    allow_partial_pushdown = "false",
+    elim_identity = "true"
+)]
 pub fn push_down_index(
     index: Index,
-    traversal: Option<IterativeMatcherRc>,
+    traversal: IterativeMatcherRc,
     suffix: Option<String>,
-    allow_partial_pushdown: Option<bool>,
-    elim_identity: Option<bool>,
+    allow_partial_pushdown: bool,
+    elim_identity: bool,
 ) -> Result<CircuitRc> {
-    let allow_partial_pushdown = allow_partial_pushdown.unwrap_or(false);
-    let elim_identity = elim_identity.unwrap_or(true);
-
     #[apply(cached_lambda)]
     #[key((index.info().hash, matcher.clone()), (HashBytes, IterativeMatcherRc))]
     #[use_try]
@@ -55,5 +57,5 @@ pub fn push_down_index(
         )
     }
 
-    push_down_rec(index, traversal.unwrap_or(default_index_traversal()))
+    push_down_rec(index, traversal)
 }
