@@ -23,6 +23,7 @@ pub struct PyUtils {
     un_flat_concat: PyObject,
     tensor_scale: PyObject,
     pub generalfunctions: std::collections::HashMap<String, PyObject>,
+    pub gen_index_function: PyObject,
     pub none: PyObject,
     einsum: PyObject,
     make_diagonal: PyObject,
@@ -31,6 +32,7 @@ pub struct PyUtils {
     pub conv: PyObject,
     pub tensor_to_bytes: PyObject,
     pub tensor_from_bytes: PyObject,
+    pub assert_tensors_close: PyObject,
 }
 
 /// misc python utilities
@@ -62,6 +64,11 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
             .unwrap()
             .extract()
             .unwrap(),
+        gen_index_function: utils
+            .getattr("gen_index_function")
+            .unwrap()
+            .extract()
+            .unwrap(),
         none: get("none"),
         einsum: get("einsum"),
         make_diagonal: get("make_diagonal"),
@@ -70,6 +77,7 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
         conv: get("conv"),
         tensor_to_bytes: get("tensor_to_bytes"),
         tensor_from_bytes: get("tensor_from_bytes"),
+        assert_tensors_close: get("assert_tensors_close"),
     }
 });
 
@@ -106,6 +114,13 @@ macro_rules! make_py_func {
     };
 }
 
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn assert_tensors_close(a: Tensor, b: Tensor) -> Result<()> {
+    PY_UTILS.assert_tensors_close.call(py, (a, b), None)?;
+    Ok(())
+}
+
 #[apply(make_py_func)]
 #[py_ident(py)]
 pub fn scalar_to_tensor(v: f64, shape: Shape, device_dtype: TorchDeviceDtype) -> Result<Tensor> {
@@ -202,10 +217,13 @@ pub static HASH_TENSOR: GILLazyPy<(PyObject, PyObject)> = GILLazyPy::new_py(|py|
 #[macro_export]
 macro_rules! pycall {
     ($f:expr,$args:expr) => {
-        Python::with_gil(|py| $f.call(py, $args, None).unwrap().extract(py).unwrap())
+        pyo3::Python::with_gil(|py| $f.call(py, $args, None).unwrap().extract(py).unwrap())
+    };
+    ($f:expr,$args:expr,raw) => {
+        pyo3::Python::with_gil(|py| $f.call(py, $args, None).unwrap())
     };
     ($f:expr, $args:expr, anyhow) => {
-        Python::with_gil(|py| {
+        pyo3::Python::with_gil(|py| {
             $f.call(py, $args, None)
                 .map_err(|z| anyhow::Error::from(z))?
                 .extract(py)
@@ -214,6 +232,24 @@ macro_rules! pycall {
     };
 }
 
+#[macro_export]
+macro_rules! atr {
+    ($x:expr,$name:ident) => {
+        Python::with_gil(|py| {
+            $x.getattr(py, stringify!($name))
+                .unwrap()
+                .extract(py)
+                .unwrap()
+        })
+    };
+    ($x:expr,$name:ident,raw) => {
+        Python::with_gil(|py| $x.getattr(py, stringify!($name)).unwrap())
+    };
+    ($x:expr,$name:ident, anyhow) => {
+        Python::with_gil(|py| Ok($x.getattr(py, stringify!($name))?.extract(py)?))
+    };
+}
+
 macro_rules! generate_extra_py_ops {
     [$($op:ident),*] => {
         paste::paste! {
@@ -240,7 +276,7 @@ macro_rules! generate_extra_py_ops {
             ///
             /// ```
             /// # use pyo3::prelude::*;
-            /// # use rust_circuit::py_types::ExtraPySelfOps;
+            /// # use rr_util::py_types::ExtraPySelfOps;
             ///
             /// #[derive(Clone, Debug, FromPyObject)]
             /// struct WrapInt(i64);
@@ -270,12 +306,12 @@ macro_rules! generate_extra_py_ops {
                 for<'a> Self: FromPyObject<'a>,
             {
                 $(
-                    fn [<py_ $op>]<'a>(self, py: Python<'a>, x: impl IntoPy<PyObject>) -> PyResult<Self> {
+                    fn [<py_ $op>](self, py: Python<'_>, x: impl IntoPy<PyObject>) -> PyResult<Self> {
                         PY_OPERATORS.$op.call1(py, (self, x))?.extract(py)
                     }
 
                     // not sure if this method should exist
-                    fn [<py_ $op _acquire>]<'a>(self, x: impl IntoPy<PyObject>) -> PyResult<Self> {
+                    fn [<py_ $op _acquire>](self, x: impl IntoPy<PyObject>) -> PyResult<Self> {
                         Python::with_gil(|py| self.[<py_ $op>](py, x))
                     }
                 )*
@@ -505,3 +541,27 @@ pub static PY_CIRCUIT_ITEMS: GILLazyPy<PyCircuitItems> = GILLazyPy::new_py(|py|
         circ_compiler_util: get("circ_compiler_util"),
     }
 });
+
+#[pyclass]
+#[derive(Copy, Clone)]
+pub struct NotSet;
+
+#[derive(FromPyObject, Clone)]
+pub enum MaybeNotSetImpl<T> {
+    NotSet(NotSet),
+    Val(T),
+}
+
+#[derive(Default, Clone)]
+pub struct MaybeNotSet<T>(pub Option<T>);
+
+impl<'source, T: FromPyObject<'source>> FromPyObject<'source> for MaybeNotSet<T> {
+    fn extract(inp: &'source PyAny) -> PyResult<Self> {
+        let x: MaybeNotSetImpl<T> = inp.extract()?;
+        let out = match x {
+            MaybeNotSetImpl::Val(x) => Some(x),
+            MaybeNotSetImpl::NotSet(_) => None,
+        };
+        Ok(MaybeNotSet(out))
+    }
+}
