@@ -466,6 +466,26 @@ impl ToPyObject for TensorAxisIndex {
 #[derive(Debug, Clone, FromPyObject)]
 pub struct TensorIndex(pub Vec<TensorAxisIndex>);
 
+// TensorIndexSync is like TensorIndex but allows tensors to have any rank
+// this is for extracting these numpy-style indices
+// we don't have full support for this, only used in new_synchronized_to_start
+pub struct TensorIndexSync(pub TensorIndex);
+impl<'source> FromPyObject<'source> for TensorIndexSync {
+    fn extract(ob: &'source PyAny) -> PyResult<Self> {
+        Ok(TensorIndexSync(TensorIndex(
+            ob.iter()?
+                .map(|x| {
+                    let x_unwrapped = x?;
+                    let tensory: PyResult<Tensor> = x_unwrapped.extract();
+                    tensory
+                        .map(|z| TensorAxisIndex::Tensor(IndexTensor(z)))
+                        .or_else(move |_| x_unwrapped.extract())
+                })
+                .collect::<PyResult<Vec<_>>>()?,
+        )))
+    }
+}
+
 impl Display for TensorIndex {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         let mut result = "[".to_owned();
@@ -631,7 +651,7 @@ impl TensorIndex {
         hasher.finalize().into()
     }
 
-    pub fn repr_bijection(&self) -> String {
+    pub fn repr_bijection(&self) -> Result<String> {
         let mut result = "[".to_owned();
         for (i, idx) in self.0.iter().enumerate() {
             result.push_str(&match idx {
@@ -642,7 +662,7 @@ impl TensorIndex {
                         + &slice.stop.map(|i| i.to_string()).unwrap_or("".to_owned())
                 }
                 TensorAxisIndex::Tensor(tensor) => {
-                    save_tensor((**tensor).clone(), false).unwrap();
+                    save_tensor((**tensor).clone(), false)?;
                     "t".to_owned() + &base16::encode_lower(&tensor.hash().unwrap())[..10]
                 }
             });
@@ -651,7 +671,7 @@ impl TensorIndex {
             }
         }
         result.push(']');
-        result
+        Ok(result)
     }
     pub fn from_bijection_string(
         string: &str,
@@ -946,6 +966,12 @@ impl TorchDeviceDtypeOp {
     }
 }
 
+impl Default for TorchDeviceDtype {
+    fn default() -> Self {
+        TorchDeviceDtypeOp::NONE.unwrap_or_defaults()
+    }
+}
+
 #[apply(python_error_exception)]
 #[base_error_name(Parse)]
 #[base_exception(PyValueError)]
