@@ -7,7 +7,7 @@ from rust_circuit import Einsum, ScalarConstant, ArrayConstant, Symbol, Add, Rea
 
 # %%
 e = Einsum.from_einsum_string("a->", ScalarConstant(0.2, (2,)))
-e.compiler_print()
+e.print()
 
 x = Einsum(
     (ScalarConstant(0.2), ()),
