@@ -28,21 +28,21 @@ base_add = Add(
 )
 
 nested_add = Add(base_add, base_add)
-nested_add.compiler_print()
+nested_add.print()
 flattened = add_flatten_once(nested_add)
 assert flattened is not None
-flattened.compiler_print()
+flattened.print()
 
 deduped = add_deduplicate(flattened)
 assert deduped is not None
-deduped.compiler_print()
+deduped.print()
 deduped_2 = add_deduplicate(base_add)
 assert deduped_2 is not None
-deduped_2.compiler_print()
+deduped_2.print()
 
 elimed = remove_add_few_input(deduped)
 assert elimed is not None
-elimed.compiler_print()
+elimed.print()
 
 scalar_merged = add_collapse_scalar_inputs(
     Add(
@@ -51,7 +51,7 @@ scalar_merged = add_collapse_scalar_inputs(
     )
 )
 assert scalar_merged is not None
-scalar_merged.compiler_print()
+scalar_merged.print()
 
 scalar_merged_call = add_collapse_scalar_inputs(
     Add(
@@ -60,7 +60,7 @@ scalar_merged_call = add_collapse_scalar_inputs(
     )
 )
 assert scalar_merged_call is not None
-scalar_merged_call.compiler_print()
+scalar_merged_call.print()
 
 add_with_rearrange = Add(
     Rearrange(ScalarConstant(2), RearrangeSpec([], [[0], [1]], [2, 3])),
@@ -68,16 +68,16 @@ add_with_rearrange = Add(
 )
 post_rearrange = add_pull_removable_axes(add_with_rearrange, True)
 assert post_rearrange is not None
-post_rearrange.compiler_print()
+post_rearrange.print()
 # %%
 
 ein = Einsum((ScalarConstant(2, (2, 3)), (0, 1)), (ScalarConstant(3, (3, 4)), (1, 2)), out_axes=(0, 2))
-ein.compiler_print()
+ein.print()
 ein_deep = Einsum((ein, (0, 1)), (ein, (0, 1)), out_axes=(0,))
-ein_deep.compiler_print()
+ein_deep.print()
 ein_flat = einsum_flatten_once(ein_deep)
 assert ein_flat is not None
-ein_flat.compiler_print()
+ein_flat.print()
 
 # %%
 
@@ -85,7 +85,7 @@ for_distribute = Einsum((base_add, ()), (base_add, ()), out_axes=())
 
 distributed = distribute_all(for_distribute)
 assert distributed is not None
-distributed.compiler_print()
+distributed.print()
 
 for_pull = Einsum(
     (Rearrange(ScalarConstant(2), RearrangeSpec([], [[0], [1]], [2, 3])), (0, 1)),
@@ -94,7 +94,7 @@ for_pull = Einsum(
 )
 pulled = einsum_pull_removable_axes(for_pull)
 assert pulled is not None
-pulled.compiler_print()
+pulled.print()
 
 rearrange_identity = Rearrange(ScalarConstant(2, (2,)), RearrangeSpec([[0]], [[0]], [2]))
 rearrange_identity_2 = Rearrange(ScalarConstant(2, (1, 1)), RearrangeSpec([[0], [1]], [[1], [0]], [1, 1]))
