@@ -29,12 +29,13 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
             add_collapse_scalar_inputs, add_deduplicate, add_elim_zeros, add_flatten_once,
             add_fuse_scalar_multiples, add_make_broadcasts_explicit, add_pull_removable_axes,
             concat_elim_identity, concat_merge_uniform, concat_pull_removable_axes,
-            concat_repeat_to_rearrange, distribute, distribute_all, einsum_concat_to_add,
+            concat_repeat_to_rearrange, distribute_all, distribute_once, einsum_concat_to_add,
             einsum_elim_identity, einsum_elim_zero, einsum_flatten_once, einsum_merge_scalars,
             einsum_nest_path, einsum_of_permute_merge, einsum_pull_removable_axes, extract_add,
             generalfunction_pull_removable_axes, index_elim_identity, index_fuse,
-            index_merge_scalar, index_split_axes, permute_of_einsum_merge, push_down_index_once,
-            rearrange_elim_identity, rearrange_fuse, rearrange_merge_scalar, remove_add_few_input,
+            index_merge_scalar, index_split_axes, make_broadcast_py, permute_of_einsum_merge,
+            push_down_index_once, rearrange_elim_identity, rearrange_fuse, rearrange_merge_scalar,
+            remove_add_few_input,
         },
         canonicalize::{canonicalize_node_py, deep_canonicalize_py},
         circuit_manipulation::{
@@ -47,7 +48,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
             generalfunction_pull_concat, index_concat_drop_unreached, split_to_concat,
         },
         deep_rewrite::{
-            compiler_simp_py, compiler_simp_step_py, compiler_simp_until_same,
             deep_heuristic_nest_adds, deep_pull_concat, deep_pull_concat_messy,
             deep_push_down_index_raw,
         },
@@ -69,7 +69,19 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit_base::PyCircuitBase>()?;
 
     m.add_class::<rr_util::rearrange_spec::RearrangeSpec>()?;
-    m.add_class::<circuit_base::GeneralFunctionSpec>()?;
+    m.add_class::<circuit_base::generalfunction::GeneralFunctionShapeInfo>()?;
+    m.add_class::<circuit_base::generalfunction::GeneralFunctionSimpleSpec>()?;
+    m.add_function(wrap_pyfunction!(
+        circuit_base::generalfunction::get_shape_info_simple,
+        m
+    )?)?;
+    circuit_base::generalfunction::register(py, m)?;
+    m.add_class::<circuit_base::generalfunction::GeneralFunctionIndexSpec>()?;
+    m.add_class::<circuit_base::generalfunction::GeneralFunctionSpecTester>()?;
+    m.add(
+        "GeneralFunctionSpecBase",
+        &*circuit_base::generalfunction::PY_WRAP_BASE,
+    )?;
     m.add_class::<rr_util::opt_einsum::EinsumSpec>()?;
     m.add_class::<circuit_base::Einsum>()?;
     m.add_class::<circuit_base::ArrayConstant>()?;
@@ -88,8 +100,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<rr_util::tensor_util::TorchDeviceDtype>()?;
     m.add_class::<rr_util::tensor_util::TorchDeviceDtypeOp>()?;
 
-    m.add_class::<circuit_base::PyGFSpecShapeGetter>()?;
-
     m.add_class::<circuit_rewrites::scheduled_execution::Schedule>()?;
     m.add_class::<circuit_rewrites::scheduled_execution::ScheduleStats>()?;
     m.add_class::<circuit_rewrites::schedule_send::ScheduleToSend>()?;
@@ -98,7 +108,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit_base::module_nodes::ModuleNodeSpec>()?;
     m.add_class::<circuit_base::module_nodes::ModuleNodeArgSpec>()?;
     m.add_class::<circuit_base::module_nodes::ModuleNodeArgSpec>()?;
-    m.add_class::<circuit_base::AutoTag>()?;
+    m.add_class::<circuit_base::Tag>()?;
     m.add_class::<circuit_base::DiscreteVar>()?;
     m.add_class::<circuit_base::StoredCumulantVar>()?;
     m.add_class::<circuit_base::Cumulant>()?;
@@ -112,9 +122,18 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<get_update_node::Getter>()?;
     m.add_class::<get_update_node::BoundGetter>()?;
     m.add_class::<get_update_node::Batcher>()?;
+    m.add_class::<circuit_base::print::PrintOptions>()?;
+    m.add_function(wrap_pyfunction!(
+        circuit_base::print::set_debug_print_options,
+        m
+    )?)?;
 
     m.add_class::<rr_util::lru_cache::TensorCacheRrfs>()?;
 
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::module_rewrite::fuse_concat_modules,
+        m
+    )?)?;
     m.add_function(wrap_pyfunction!(circuit_base::circuit_is_leaf_py, m)?)?;
     m.add_function(wrap_pyfunction!(
         circuit_base::circuit_is_irreducible_node_py,
@@ -158,7 +177,8 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(concat_pull_removable_axes, m)?)?;
     m.add_function(wrap_pyfunction!(einsum_pull_removable_axes, m)?)?;
     m.add_function(wrap_pyfunction!(add_make_broadcasts_explicit, m)?)?;
-    m.add_function(wrap_pyfunction!(distribute, m)?)?;
+    m.add_function(wrap_pyfunction!(make_broadcast_py, m)?)?;
+    m.add_function(wrap_pyfunction!(distribute_once, m)?)?;
     m.add_function(wrap_pyfunction!(distribute_all, m)?)?;
     m.add_function(wrap_pyfunction!(einsum_of_permute_merge, m)?)?;
     m.add_function(wrap_pyfunction!(permute_of_einsum_merge, m)?)?;
@@ -168,9 +188,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(index_split_axes, m)?)?;
     m.add_function(wrap_pyfunction!(add_elim_zeros, m)?)?;
 
-    m.add_function(wrap_pyfunction!(compiler_simp_py, m)?)?;
-    m.add_function(wrap_pyfunction!(compiler_simp_step_py, m)?)?;
-    m.add_function(wrap_pyfunction!(compiler_simp_until_same, m)?)?;
     m.add_function(wrap_pyfunction!(deep_canonicalize_py, m)?)?;
     m.add_function(wrap_pyfunction!(canonicalize_node_py, m)?)?;
 
@@ -268,10 +285,16 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
 
+    m.add_class::<circuit_rewrites::deep_rewrite::SimpFnSubset>()?;
     m.add_function(wrap_pyfunction!(
-        circuit_rewrites::deep_rewrite::basic_simp,
+        circuit_rewrites::deep_rewrite::compiler_simp_step_py,
         m
     )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::deep_rewrite::compiler_simp_py,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(circuit_rewrites::deep_rewrite::simp, m)?)?;
     m.add_function(wrap_pyfunction!(circuit_base::deep_map_preorder_py, m)?)?;
     m.add_function(wrap_pyfunction!(circuit_base::deep_map_py, m)?)?;
 
@@ -317,6 +340,18 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         get_update_node::sample_transform::default_var_matcher,
         m
     )?)?;
+    m.add_function(wrap_pyfunction!(
+        get_update_node::library::replace_outside_traversal_symbols_py,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        nb_operations::modules::extract_rewrite,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        get_update_node::library::apply_in_traversal_py,
+        m
+    )?)?;
     m.add_class::<get_update_node::sample_transform::RandomSampleSpec>()?;
     m.add_class::<get_update_node::sample_transform::RunDiscreteVarAllSpec>()?;
     m.add_class::<get_update_node::sample_transform::SampleTransform>()?;
@@ -332,7 +367,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
     m.add_function(wrap_pyfunction!(
-        circuit_base::parsing::parse_compiler_repr_bijection_py,
+        circuit_base::parsing::parse_circuit_py,
         m
     )?)?;
 
@@ -342,6 +377,8 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
     m.add_function(wrap_pyfunction!(rr_util::py_types::hash_tensor, m)?)?;
+    m.add_class::<rr_util::py_types::NotSet>()?;
+    m.add("NOT_SET", rr_util::py_types::NotSet)?;
 
     m.add_function(wrap_pyfunction!(rr_util::tensor_db::save_tensor, m)?)?;
     m.add_function(wrap_pyfunction!(rr_util::tensor_db::get_tensor_prefix, m)?)?;
@@ -353,26 +390,8 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         rr_util::tensor_db::migrate_tensors_from_old_dir,
         m
     )?)?;
-    m.add_class::<nb_operations::nest_einsums::NestEinsumsRest>()?;
-    m.add_class::<nb_operations::nest_einsums::NestEinsumsSpecInfo>()?;
-    m.add_class::<nb_operations::nest_einsums::NestEinsumsIdxsInfo>()?;
 
-    m.add_function(wrap_pyfunction!(
-        nb_operations::nest_einsums::einsum_flatten,
-        m
-    )?)?;
-    m.add_function(wrap_pyfunction!(
-        nb_operations::nest_einsums::nest_einsums,
-        m
-    )?)?;
-    m.add_function(wrap_pyfunction!(
-        nb_operations::nest_einsums::nest_einsums_strict,
-        m
-    )?)?;
-    m.add_function(wrap_pyfunction!(
-        nb_operations::nest_einsums::nest_flat_einsum_strict,
-        m
-    )?)?;
+    nb_operations::nest::register(py, m)?;
 
     m.add_function(wrap_pyfunction!(
         nb_operations::index_rewrites::default_index_traversal,
@@ -383,6 +402,15 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
 
+    m.add_function(wrap_pyfunction!(
+        nb_operations::distribute_and_factor::traverse_until_depth,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        nb_operations::distribute_and_factor::distribute,
+        m
+    )?)?;
+
     // create dummy object for all the type aliases we use in the stub file
     // TODO: maybe these should somehow be the actual types we use in the stub file
     // (this is kinda annoying)
@@ -402,7 +430,10 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         "NestEinsumsSpecMultiple",
         "NestEinsumsSpecSub",
         "NestEinsumsSpec",
-        "EinsumNestIdxsItem",
+        "NestAddsSpecMultiple",
+        "NestAddsSpecSub",
+        "NestAddsSpec",
+        "GeneralFunctionSpec",
     ] {
         m.add(dummy_name, py.None())?;
     }
