@@ -1,6 +1,6 @@
 use std::fs;
 
-use circuit_base::parsing::parse_compiler_repr_bijection;
+use circuit_base::parsing::parse_circuit;
 use circuit_rewrites::circuit_optimizer::{
     optimize_circuit, OptimizationContext, OptimizationSettings,
 };
@@ -16,7 +16,7 @@ fn benchmarking_circs() {
         let circuits: Vec<_> = fs::read_dir(format!("{}/ryan/compiler_benches/", get_rrfs_dir()))
             .unwrap()
             .filter_map(|d| {
-                parse_compiler_repr_bijection(
+                parse_circuit(
                     &fs::read_to_string(d.unwrap().path()).unwrap(),
                     Default::default(),
                     Default::default(),
